#include "AnalysisService.h"
#include "AnalysisAlgorithmMap.h"
#include "AnalysisAlgorithm.h"
#include "AnalysisWorkerGroup.h"
#include "DcsInterface.h"
#include "ConfigurationInterface.h"

#include "Sct/SctNames.h"
#include "Sct/IpcObjectException.h"
#include "Sct/IS/IOManagerIS.h"
#include "Sct/IS/IONameIS.h"
#include "Sct/ConfigurationException.h"
#include "sctConf/configuration.h"
#include "SctData/TestResult.h"
#include "sctConfIPC/configipc.h"
#include "CalibrationController/IS/TestData.h"

#include <ipc/core.h>
#include <boost/scoped_ptr.hpp>

using namespace Sct;
using namespace Sct::IS;
using namespace SctData;
using boost::scoped_ptr;

namespace SctAnalysis {

AnalysisService::AnalysisService(const string& fitStrategyName) : 
  IPCNamedObject<POA_AnalysisServiceI::AnalysisServiceInterface>(
							     SctNames::getPartition(),
							     AnalysisServiceI::AnalysisServiceInterface::instanceName),
  dcsinterface(0),
  infoReceiver(new ISInfoReceiver(SctNames::getPartition())) {
    if (!infoReceiver.get())
        throw ConfigurationException("AnalysisService::AnalysisService can't make infoReceiver ", __FILE__, __LINE__) ;
    std::cout << "Making DcsInterface ... " << std::endl;
    boost::shared_ptr<SctConfiguration::Configuration> configipc(new SctConfiguration::ConfigIPC);
    boost::shared_ptr<ConfigurationInterface> config_local (new ConfigurationInterface(configipc));
    dcsinterface = new DcsInterface(config_local);

    try{
      config_local->update();
    }catch(Sct::Throwable& e){
      e.sendToMrs(MRS_INFORMATION);
    }catch(SctConfiguration::ConfigurationException &e){
      Sct::ConfigurationException(e.what(), __FILE__, __LINE__).sendToMrs(MRS_INFORMATION);
    }
    setFitStrategy(fitStrategyName);
    workergroup = new AnalysisWorkerGroup();
}

IPCServer& AnalysisService::getServer() throw() {
    static IPCServer server;
    return server;
}

void AnalysisService::run() {
    // start up worker thread.
    workergroup->go(1);

    // subscribe to various IS stuff.
    infoReceiver->subscribe(SctNames::getControlDataName().c_str(), ".*TestData.*", testDataCallback, this);
    infoReceiver->subscribe(SctNames::getFittedDataName().c_str(),  ".*FitScanResult.*", scanResultCallback, this);
    infoReceiver->subscribe(SctNames::getEventDataName().c_str(), ".*RawScanResult.*", scanResultCallback, this);
}

  CORBA::Short AnalysisService::busy() {
    boost::recursive_mutex::scoped_lock lock(m_status_access);
    return workergroup->busy();
  }

  CORBA::Short AnalysisService::queueLength() {
  boost::recursive_mutex::scoped_lock lock(m_status_access);
    return workergroup->busy();
  }

char* AnalysisService::status() throw() {
  boost::recursive_mutex::scoped_lock lock(m_status_access);
    // Have to make a new char* because IPC wants to delete it after use!
    ostringstream os;
    workergroup->printStatus(os);
    os << "\n" << AnalysisAlgorithmMap::instance().getAllStatus() << endl;
    if (dcsinterface){
      os << "AnalysisService : DCS INFORMATION" << endl;
      dcsinterface->printStatus(os);
    }

    const unsigned length = os.str().length()+1;
    char *statusMessage = new char[length];
    strcpy(statusMessage, os.str().c_str());
    return statusMessage;
}

void AnalysisService::analyzeModule(const char* testname, const char* modulename) throw() {
    try {
        shared_ptr<TestData> testdata( new TestData() );
        ISInfoDictionary& id = SctNames::getISDictionary();
        ISInfo::Status result = id.findValue(testname, *testdata);

        if (result != ISInfo::Success) {
            string os = "Error reading from IS server.  Couldn't get: ";
            os += testname;
            throw IsException(result, os, __FILE__, __LINE__);
        }
        cout << "read " << testname << " which has #scans=" <<  testdata->nScans << endl;
        if (workergroup->findTest(*testdata).get() == 0 ) {
            workergroup->addTest(testdata);
        }

        for (unsigned iscan=testdata->startScanNumber;
                iscan<(testdata->startScanNumber + testdata->nScans);
                ++iscan ) {
            {
                ostringstream name_raw;
		///@todo This seems broken to me.
                name_raw << "SctData::RawScanResult." <<  testdata->runNumber << "." << iscan << "." << modulename;
                cout << "Looking for " << name_raw.str() << endl;
                ISInfoIterator rawIter(SctNames::getPartition(), SctNames::getEventDataName().c_str(), name_raw.str().c_str());
                while ( rawIter() ) {
                    try {
                        cout << "trying to add " << rawIter.name() << endl;
                        workergroup->push(shared_ptr<IOName>(new IONameIS(rawIter.name())));
                    } catch ( Sct::Throwable& e) {
                        e.sendToMrs(MRS_ERROR);
                    }
                }
            } {
                ostringstream name_fit;
		///@todo This seems broken to me.
                name_fit << "SctData::FitScanResult." <<  testdata->runNumber << "." << iscan << "." << modulename;
                cout << "Looking for " << name_fit.str() << endl;
                ISInfoIterator fitIter(SctNames::getPartition(), SctNames::getFittedDataName().c_str(), name_fit.str().c_str());
                while ( fitIter() ) {
                    try {
                        cout << "trying to add " << fitIter.name() << endl;
                        workergroup->push(shared_ptr<IOName>(new IONameIS(fitIter.name())));
                    } catch ( Sct::Throwable& e) {
                        e.sendToMrs(MRS_ERROR);
                    }
                }
            }
        }
    } catch (Throwable& e) {
        e.sendToMrs(MRS_ERROR);
        return;
    }
}

void AnalysisService::purge() throw() {
    try {
        cout << "Doing purge" << endl;
        workergroup->purge();
        cout << "Memory purge completed" << endl;
    } catch (Throwable& e) {
        e.sendToMrs(MRS_ERROR);
        return;
    }
}

void AnalysisService::analyze(const char* name) throw() {
    analyzeModule(name, "*");
}

void AnalysisService::scanResultCallback(ISCallbackInfo * isc) {
    try {
        if (isc->reason() != ISInfoCreated )
            return;
        /** push the IS name of the ScanResult onto the queue */
        cout << "AnalysisService scanResultCallback on " << string(isc->name())<<endl;
        //cout << "AnalysisService scanResultCallback busy=" << AnalysisWorkerGroupFactory::pointer()->busy()
        //<< " Q=" << AnalysisWorkerGroupFactory::pointer()->queueSize()
        //<< " Nworker=" << AnalysisWorkerGroupFactory::pointer()->nWorkers() << endl;
        instance().workergroup->push(shared_ptr<IOName>(new IONameIS(isc->name())));
        //cout << " scanResultCallback done " << endl;
    } catch(Sct::Throwable& e) {
        e.sendToMrs(MRS_ERROR);
    }
}

void AnalysisService::testDataCallback(ISCallbackInfo * isc) {
    try {
        shared_ptr<TestData> testdata( new TestData() );
        isc->value(*testdata);

        if (isc->reason() == ISInfoCreated ) {
            instance().workergroup->push(shared_ptr<IOName>(new IONameIS(isc->name())));
	} else if (isc->reason() == ISInfoUpdated && testdata->status == TestData::ABORTED) {
	    instance().workergroup->removeTestsUpTo(testdata);
	    
	} else if (isc->reason() == ISInfoDeleted ) {
            instance().workergroup->removeTestsUpTo(testdata);
        }
    } catch (Sct::Throwable& e) {
        e.sendToMrs(MRS_ERROR);
    }
}

void AnalysisService::setFitStrategy(const string& name) throw(LogicError) {
    fitStrategy = SctFitter::FitStrategyFactory::instance().getStrategy(name);
}

SctFitter::FitStrategy& AnalysisService::getFitStrategy() const throw(LogicError) {
    if (!fitStrategy)
        throw InvariantViolatedError("Fitter::getStrategy() no fit strategy defined", __FILE__, __LINE__);
    return *fitStrategy;
}

AnalysisService& AnalysisService::instance() {
    if (!service)
        return initialize();
    return *service;
}

AnalysisService& AnalysisService::initialize(const string& fitStrategyName) {
    if (service) {
        service->setFitStrategy(fitStrategyName);
    } else {
        service = new AnalysisService(fitStrategyName);
    }
    return *service;
}

DcsInterface& AnalysisService::getDcsInterface(){
  if (!dcsinterface){
    throw IllegalStateError("Cant return dcsinterface",__FILE__, __LINE__);
  }else{
    return *dcsinterface;
  }
}


AnalysisService* AnalysisService::service = 0;

}// end of namespace SctAnalysis
