#include "Fitter.h"
#include "../src/FitStrategy.h"
#include "FitterWorkerGroup.h"
#include "ThresholdFitAlgorithm.h"
#include "ThresholdFitAlgorithm2.h"

#include "Sct/SctNames.h"
#include "Sct/SctParameters.h"
#include "Sct/DestroyingDeleter.h"

#include <iostream>
#include <vector>
#include <cassert>
#include <ctime>
#include <TMath.h>
#include <boost/timer.hpp>

using namespace std;
using namespace Sct;
using namespace SctData;
using namespace boost;

namespace SctFitter {
void doFits(ISCallbackInfo * isc);

boost::shared_ptr<Fitter> Fitter::fitter = boost::shared_ptr<Fitter>();

Fitter& Fitter::instance() {
    if (!(fitter.get()))
        return initialize();
    return *fitter;
}

Fitter& Fitter::initialize(const string& fitStrategyName) throw(ConfigurationException) {
    if (fitter.get()) {
        fitter->internal_setFitStrategy(fitStrategyName);
        return *fitter;
    }
    fitter = ::Sct::DestroyingDeleter<Fitter>::shared_ptr(new Fitter(fitStrategyName));
    return *fitter;
}

Fitter::Fitter(const string& fitStrategyName) throw(ConfigurationException) : 


  IPCNamedObject<POA_FitterI::FitterInterface,
		 ipc::multi_thread>(

				    SctNames::getPartition(),
				    FitterI::FitterInterface::instanceName)


					  //LESTER: Was IPCObject(FitterI_C_FitterInterface_instanceName,
					  //&getFitterServer())
  ,m_nFitsDone(0), m_nFitErrors(0), m_nScans(0), m_scanTimeTaken(0), m_ioTimeTaken(0), m_scan("No last scan"), m_fitStrategy(0) {

    auto_ptr<ISInfoReceiver> ir(new ISInfoReceiver(SctNames::getPartition()));
    if (!ir.get())
        throw ConfigurationException("Fitter::initialize can't make infoReceiver ", __FILE__, __LINE__) ;
    m_infoReceiver  = ir;

    internal_setFitStrategy(fitStrategyName);
    workergroup = new FitterWorkerGroup();
}

Fitter::~Fitter() throw() {}


IPCServer& Fitter::getFitterServer() throw() {
  static IPCServer fitterServer;
    return fitterServer;
}



void Fitter::go(unsigned nWorker) throw(IsException) {
    ///@todo proper RegExp
    ISInfo::Status s=m_infoReceiver->subscribe(SctNames::getEventDataName().c_str(),"SctData::RawScanResult.*", doFits);
    if (s!=ISInfo::Success) {
        ostringstream os;
        os <<"Fitter::go() Could not subscribe to "<<SctNames::getEventDataName();
        throw IsException(s, os.str(), __FILE__, __LINE__);
    }
    workergroup->go(nWorker);

    //Not done since Fitter is an IPC server, and so IPCServer.run() is called elsewhere.
    //infoReceiver->run();
}

// The infoReceiver callback function must take a static function as an argument:
void Fitter::doFits(ISCallbackInfo * isc) throw(IsException, LogicError) {
    // find out why the callback function has been called:
    if (isc->reason() != ISInfoCreated && isc->reason() != ISInfoUpdated )
	return;
    
    // put it in the queue
    Fitter::instance().workergroup->push(isc->name());
}

  //CORBA
void Fitter::fit(const char* name) throw() {
    // put it in the queue
    workergroup->push(name);
}

  //CORBA
void Fitter::fitAll() throw() {

    ISInfoIterator iter(SctNames::getPartition(), SctNames::getEventDataName().c_str(), "SctData::RawScanResult.*");
    while (iter()) {
	// put it in the queue
	workergroup->push(iter.name());
    }
}

  //CORBA
char* Fitter::getFitOptions() throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    string options = "";
    try {
        options = internal_getFitStrategy().getOptions();		
    } catch (LogicError &e) {
	//Um don't do anything!
    }
    char *newOpt=new char[options.length()+1];
    strcpy(newOpt, options.c_str());
    return newOpt;
}

  //CORBA
char* Fitter::status() throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    // Have to make a new char* because IPC wants to delete it after use!
    string msg = internal_status();
    char *statusMessage = new char[msg.length()+1];
    strcpy(statusMessage, msg.c_str());
    return statusMessage;
}

void Fitter::internal_setFitOptions(const string & opt) throw (LogicError) {
  recursive_mutex::scoped_lock scoped_lock(counterMutex);
  internal_getFitStrategy().setOptions(opt);
}

  //CORBA
void Fitter::setFitOptions(const char *n) throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    string name( n );
    try {
        internal_setFitOptions(name);
    } catch (LogicError &e) {
        return;
    }
}

  //CORBA
void Fitter::useAnalyticAlgorithm (CORBA::Boolean use) {
    if (use) {
	ThresholdFitAlgorithm2::putInMap();
    } else {
	ThresholdFitAlgorithm::putInMap();
    }
}

bool Fitter::internal_isUsingAnalyticAlgorithm() const {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    return ThresholdFitAlgorithm2::isInMap();
}

  //CORBA
CORBA::Boolean Fitter::isUsingAnalyticAlgorithm() {
     recursive_mutex::scoped_lock scoped_lock(counterMutex);
   return internal_isUsingAnalyticAlgorithm();
}

char* Fitter::getFitStrategy() throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    string strat;
    try {
        strat = internal_getFitStrategy().getName();
    } catch (LogicError &e) {
	//Do nothing!
    }    
    char *newOpt=new char[strat.length()+1];
    strcpy(newOpt, strat.c_str());
    return newOpt;
}

  //CORBA
void Fitter::setFitStrategy(const char *n) throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    string name( n );
    cout <<"trying to set fit strategy "<<name<<endl;
    try {
        internal_setFitStrategy(name);
    } catch (LogicError& e) {
        cout <<"failed  to set fit strategy "<<name<<endl;
        return;
    }
    cout <<"set fit strategy "<<name<<endl;
}

const char* Fitter::internal_status() const throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    string strategyName;
    try {
        strategyName=internal_getFitStrategy().getName();
    } catch (LogicError &e) {
        strategyName="NONE";
    }
    ostringstream oss;
    oss << "Fitter Strategy=" << strategyName <<  " Fit Options=" << internal_getFitStrategy().getOptions()
	<< " Using analytic fits=" << internal_isUsingAnalyticAlgorithm() << endl;
    oss << "Workers=" << workergroup->nWorkers();
    oss << ", (Busy=" << workergroup->busy() << "), Fits Done=" << internal_nFitsDone() << ", Errors=";
    oss << internal_nFitErrors() << ", Scans Done=" << m_nScans << "\nLast scan: " << internal_lastScan(); 
    oss << "\n Timing.  I/O time: " << m_ioTimeTaken << " Scan time: "; 
    oss << m_scanTimeTaken << " Approx average time/scan/thread: " << getAverageTimePerScan() << endl;
    m_status = oss.str();
    return m_status.c_str();
}

void Fitter::incrementFitErrors() throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    ++m_nFitErrors;
}

void Fitter::incrementFitsDone() throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    ++m_nFitsDone;
}

void Fitter::scanDone(double time) throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    ++m_nScans;
    m_scanTimeTaken += time;
}

void Fitter::addIOTime(double time) throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    m_ioTimeTaken += time;
}


double Fitter::getAverageTimePerScan() const throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    double time = (m_ioTimeTaken+m_scanTimeTaken)/workergroup->nWorkers();
    if (m_nScans > 0) time /= m_nScans;     
    return time;
}


long Fitter::internal_nFitsDone() const throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    return m_nFitsDone;
}

long Fitter::internal_nFitErrors() const throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    return m_nFitErrors;
}

  //CORBA
long Fitter::nFitsDone() throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    return internal_nFitsDone();
}

  //CORBA
long Fitter::nFitErrors() throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    return internal_nFitErrors();
}

  //CORBA
long Fitter::queueLength() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    return workergroup->queueSize();
}

  //CORBA
long Fitter::busy() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    return workergroup->busy();
}

const char* Fitter::internal_lastScan() const throw() {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    return m_scan.c_str();
}

  //CORBA
char* Fitter::lastScan() throw() {    
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    char * message = new char[m_scan.length()+1];
    strcpy(message, m_scan.c_str() );
    return message;
}

FitStrategy& Fitter::internal_getFitStrategy() const throw(LogicError) {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    if (!m_fitStrategy)
        throw InvariantViolatedError("Fitter::getStrategy() no fit strategy defined", __FILE__, __LINE__);
    return *m_fitStrategy;
}

void Fitter::internal_setFitStrategy(const string & name) throw(LogicError) {
    recursive_mutex::scoped_lock scoped_lock(counterMutex);
    m_fitStrategy = FitStrategyFactory::instance().getStrategy(name);
}

} // end of namespace SctFitter
