/*
 * SCTInterface.java
 *
 * Created on 15 November 2002, 12:01
 */

package GuiComponents.System;

import ipc.*;
import java.util.*;
import java.util.logging.*;
import SctTestApi.*;
import Sct_CalibrationController.*;
import Sct_SctApi.*;
import AnalysisServiceI.*;
import ArchivingServiceI.*;
import sctConf.*;
import FitterI.*;


/**
 * This class is an interface to the rest of the Software System
 * It contains methods to get access to all the various components that
 * may or may not be running.
 *
 * Users should probably not cache references for long periods of time
 *
 * This class and ISInterface mirror some of the methods from SctNames in the C++ code
 *
 * @author Matthew Palmer
 */
public class SystemInterface {    
    
    /**
     * Get the singleton interface.
     */
    public static SystemInterface getInstance() {
        return instance;
    }

    /**
     * Re-obtain all connections to external processes.
     */
    public void refresh() {
        p = null;               
        testconfig = null;
        highlevel = null;
        cc = null; 
        api = null;
        config = null;
        analysis = null;
	archiving = null;
        fitter = null;
    }

    /**
     * Get the IPC Partition that all our processes are contained in.
     */
    public Partition getPartition() {
	if (p == null) {
	    try {
		p = new Partition(partitionName); 
	    } catch (PartitionNotFoundException e) {
		logger.info("Failed to get partition: " + partitionName + ": " + e.getMessage());
	    }
	}
        return p;
    }

    /**
     * Get a reference to the SctTest Configuration interface
     */
    public SctTestApi.Configuration getTestConfiguration() {
        if (testconfig == null) {
            try {
                org.omg.CORBA.Object ob = getPartition().lookup(sctConf.Configuration.class, SctTestApi.Configuration.instanceName);
                testconfig = SctTestApi.ConfigurationHelper.narrow(ob);
		if (testconfig == null) {
		    logger.info("Failed to lookup SctTest::Configuration using: " + SctTestApi.Configuration.instanceName);
		}
            } catch (org.omg.CORBA.SystemException se) {
		logger.info("Failed to lookup SctTest::Configuration using: " + SctTestApi.Configuration.instanceName + ". Exception: " + se.getMessage());
		//} catch (ClassNotFoundException ce) {
		//logger.info("Failed to lookup Class name for sctConf.Configuration!. Exception: " + ce.getMessage());
	    } catch (NullPointerException npe) {
		//Do nothing - this must have been caused by not being able to get the partition, which we've already logged
	    }
        }
        return testconfig;
    }

    /**
     * Get a reference to the SctTest HighLevelApi interface
     */
    public HighLevelApi getTestHighLevelApi() {
        if (highlevel == null) {
            try {
                org.omg.CORBA.Object ob = getPartition().lookup(SctTestApi.HighLevelApi.class, HighLevelApi.instanceName);
                highlevel = HighLevelApiHelper.narrow(ob);
		if (highlevel == null) {
		    logger.info("Failed to lookup SctTest::HighLevelApi using: " + HighLevelApi.instanceName);
		}
            } catch (org.omg.CORBA.SystemException se) {
		logger.info("Failed to lookup SctTest::HighLevelApi using: " + HighLevelApi.instanceName + ". Exception: " + se.getMessage());
		//} catch (ClassNotFoundException ce) {
		//logger.info("Failed to lookup Class name for HighLevelApi.HighLevelApi!. Exception: " + ce.getMessage());
	    } catch (NullPointerException npe) {
		//Do nothing - this must have been caused by not being able to get the partition, which we've already logged
	    }
        }
        return highlevel;
    }


    /**
     * Gets a reference to the CalibrationController
     */
    public CalibrationController getCalibrationController() {
        if (cc == null) {
            try {
                org.omg.CORBA.Object ob = getPartition().lookup(Sct_CalibrationController.CalibrationController.class, CalibrationController.instanceName);
                cc = CalibrationControllerHelper.narrow(ob);
            } catch (org.omg.CORBA.SystemException se) {                
                logger.info("Failed to lookup CalibrationController using: " + CalibrationController.instanceName + ". Exception: " + se.getMessage());
		//} catch (ClassNotFoundException ce) {
		//logger.info("Failed to lookup Class name for CalibrationController.CalibrationController!. Exception: " + ce.getMessage());
	    } catch (NullPointerException npe) {
		//Do nothing - this must have been caused by not being able to get the partition, which we've already logged
	    }
        }
        return cc;        
    }
    
    /**
     * Gets a reference to the Api
     */
    public SctApiIPC getSctApi() {
        if (api == null) {
            try {
                org.omg.CORBA.Object ob = getPartition().lookup(Sct_SctApi.SctApiIPC.class, SctApiIPC.instanceName);
                api = SctApiIPCHelper.narrow(ob);
            } catch (org.omg.CORBA.SystemException se) {                
                logger.info("Failed to lookup AnalysisService using: " + SctApiIPC.instanceName + " . Exception: " + se.getMessage());
		//} catch (ClassNotFoundException ce) {
		//logger.info("Failed to lookup Class name for Sct_SctApi.SctApiIPC!. Exception: " + ce.getMessage());
	    } catch (NullPointerException npe) {
		//Do nothing - this must have been caused by not being able to get the partition, which we've already logged
	    }
        }
        return api;        
    }
    
    /**
     * Gets a reference to the configuration service
     */
    public sctConf.Configuration getConfigurationService() {
        if (config == null) {
            try {
                org.omg.CORBA.Object ob = getPartition().lookup(sctConf.Configuration.class, sctConf.Configuration.instanceName); 
                config = sctConf.ConfigurationHelper.narrow(ob);
            } catch (org.omg.CORBA.SystemException se) {                
                logger.info("Failed to lookup Configuration using: " + sctConf.Configuration.instanceName + ". Exception: " + se.getMessage());
		//} catch (ClassNotFoundException ce) {
		//logger.info("Failed to lookup Class name for sctConf.Configuration!. Exception: " + ce.getMessage());
	    } catch (NullPointerException npe) {
		//Do nothing - this must have been caused by not being able to get the partition, which we've already logged
	    }
        }
        return config;        
    }
    
    /**
     * Gets a reference to the analysis service
     */
    public AnalysisServiceInterface getAnalysisService() {
        if (analysis == null) {
            try {
                org.omg.CORBA.Object ob = getPartition().lookup(AnalysisServiceI.AnalysisServiceInterface.class, AnalysisServiceInterface.instanceName);
                analysis = AnalysisServiceInterfaceHelper.narrow(ob);
            } catch (org.omg.CORBA.SystemException se) {                
                logger.info("Failed to lookup AnalysisService using: " + AnalysisServiceInterface.instanceName + ". Exception: " + se.getMessage());
		//} catch (ClassNotFoundException ce) {
		//logger.info("Failed to lookup Class name for AnalysisServiceI.AnalysisServiceInterface!. Exception: " + ce.getMessage());
	    } catch (NullPointerException npe) {
		//Do nothing - this must have been caused by not being able to get the partition, which we've already logged
	    }
        }
        return analysis;                
    }
    
    /**
     * Gets a reference to the archiving service
     */
    public ArchivingServiceInterface getArchivingService() {
        if (archiving == null) {
            try {
                org.omg.CORBA.Object ob = getPartition().lookup(ArchivingServiceI.ArchivingServiceInterface.class, ArchivingServiceInterface.instanceName);
                archiving = ArchivingServiceInterfaceHelper.narrow(ob);
            } catch (org.omg.CORBA.SystemException se) {
                logger.info("Failed to lookup ArchivingService using: " + ArchivingServiceInterface.instanceName + ". Exception: " + se.getMessage());
		//} catch (ClassNotFoundException ce) {
		//logger.info("Failed to lookup Class name for ArchivingServiceInterface.ArchivingServiceInterface!. Exception: " + ce.getMessage());
            } catch (NullPointerException npe) {
                //Do nothing - this must have been caused by not being able to get the partition, which we've already logged
            }
        }
        return archiving;
    }


    /**
     * Gets a reference to the fitting service
     */
    public FitterInterface getFittingService() {
        if (fitter == null) {
            try {
                org.omg.CORBA.Object ob = getPartition().lookup(FitterI.FitterInterface.class, FitterInterface.instanceName);
                fitter = FitterInterfaceHelper.narrow(ob);
            } catch (org.omg.CORBA.SystemException se) {                
                logger.info("Failed to lookup Fitter using: " + FitterInterface.instanceName + ". Exception: " + se.getMessage());
		//} catch (ClassNotFoundException ce) {
		//logger.info("Failed to lookup Class name for FitterI.FitterInterface!. Exception: " + ce.getMessage());
	    } catch (NullPointerException npe) {
		//Do nothing - this must have been caused by not being able to get the partition, which we've already logged
	    }
        }
        return fitter;                
    }
    
    /** Creates a new instance of SCTInterface */
    private SystemInterface() {
        logger = Logger.getLogger("GuiComponents.System");     
        refresh();
    }            
    
    private static SystemInterface instance = new SystemInterface();
    private static final String partitionName = "SCT";
    private Partition p;
    private Logger logger;    
    private SctTestApi.Configuration testconfig;
    private HighLevelApi highlevel;
    private CalibrationController cc;
    private SctApiIPC api;
    private sctConf.Configuration config;
    private AnalysisServiceInterface analysis;
    private ArchivingServiceInterface archiving;
    private FitterInterface fitter;
}
