#include <unistd.h>
#include <stdlib.h>
#include <iostream>

#include <mrs/message.h>
#include <is/isinfotmpl.h>

#include <algorithm>
#include <sstream>
#include <memory>

#include "SctApi.hh"
#include "SctApiServer.h"
#include "sctConfIPC/configipc.h"
#include "SctApi/autoConfig.h"

#include "Sct/SctNames.h"
#include "Sct/OmniMarshalling.h"
#include "Sct/DestroyingDeleter.h"
#include "sctConfIPC/marshalling.h"
#include "marshalling.h"

#include "SctApi/SctApi.h"
#include "SctApi/SctApiImpl.h"

#include "TriggerWrapper.h"
#include    "ScanWrapper.h"

#include "Sct/SequenceUtils.h"
#include "Sct/StringStreamer.h"

#include <boost/date_time/posix_time/posix_time.hpp>

using namespace Sct;
using namespace std;
//using namespace boost::posix_time;

static Sct_SctApi::SctApiException myRaise( ::SctApi::SctApiException &s);

static inline SctApi::BankType myTranslateBank(Sct_SctApi::BankType bank);

/*
  CORBA::ULong is long unsigned int
 */

SctApiServer *SctApiServer::singleton = 0;

SctApiServer::SctApiServer(SctApi::SctApi& api, 
			   const char* interfaceName) : 
  IPCNamedObject<POA_Sct_SctApi::SctApiIPC>(Sct::SctNames::getPartition(),
					    interfaceName),
  scans(0), 
  defaultScan(0),
  defaultTrigger(0),
  mrsStream(/*LESTER server->partition()*/ Sct::SctNames::getPartition()), 
  isDict(/*LESTER server->partition()*/ Sct::SctNames::getPartition()),
  ddcSingleTimeout(3), ddcMultipleTimeout(15),
  api(api)
{
  publish();

  ISInfoInt isPresent(1);

  defaultScan = new ScanServer;
  api.setMrsStream(&mrsStream);
  api.setIsDictionary(&isDict);

  mrsStream << "SCTAPI_CONSTRUCT" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("Constructor called") << ENDM;

  ISInfo::Status s=isDict.contains("SCTAPIServer.present");
  if(s==ISInfo::Success) {
    std::cout << "*** Filename already in IS, will run anyway but there could be another service running?\n";
  } else {
    s=isDict.insert("SCTAPIServer.present", isPresent);
    if (s!=ISInfo::Success){
      mrsStream << "IS_ERROR" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("Error publishing presence to IS") << MRS_PARAM<int>("Status",static_cast<int>(s)) << ENDM;
    }
  }
  
  singleton = this;
}

SctApiServer::~SctApiServer() 
{
  mrsStream << "SCTAPI_DESTRUCT" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("Destructor called") << ENDM;

  isDict.remove("SCTAPIServer.present");

  singleton = 0;

  withdraw();
  //deactivate();

  std::cout << "Forcefully deleting " << scans.size() << " scans\n";
  scans.clear();
  /*
    // This old method is (hopefuly) no longer needed now that elements of "scans" have custom delters!
    for(Scans::iterator i = scans.begin();
      i!= scans.end(); 
      i++) {
    delete *i;
  }
  */

#warning LESTER SUSPECTS THAT THE AUTHORS OF THIS ROUTINE SHOULD ALSO BE FREEING THINGS LIKE THE DEFAULT SCAN AND THE DEFAULT TRIGGER!
}

void SctApiServer::initialiseAll (CORBA::ULong runNumber)
{
  try {
    mrsStream << "SCTAPI_INIT" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT(Sct::StringStreamer<< "initialiseAll called with run number "<< runNumber) << ENDM;

    worker().initialiseAll(runNumber);

    ISInfoInt present(2);

    isDict.update("SCTAPIServer.present", present);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by initialiseAll: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::shutdownAll ()
{
  try {
    mrsStream << "SCTAPI_SHUTDOWN" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("shutdownAll called") << ENDM;

    worker().shutdownAll();

    ISInfoInt present(1);

    isDict.update("SCTAPIServer.present", present);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by shutdownAll: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::status ()
{
  worker().status();
}

void SctApiServer::awaitResponse (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  try {
    worker().awaitResponse((unsigned int)partition, (unsigned int)crate, (unsigned int)rod, 5);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by awaitResponse: " << s.what() << endl;
    throw myRaise(s);
  }
}

Sct_SctApi::DataBlock * SctApiServer::getResponse (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  unsigned long length;

  unsigned long *body = worker().getResponse((unsigned int)partition, (unsigned int)crate, (unsigned int)rod, length);

  return copyArrayToCorba<Sct_SctApi::DataBlock, unsigned long>(body, length);
}

void SctApiServer::configureBOC (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  try {
    worker().configureBOC((unsigned int) partition, (unsigned int)crate, (unsigned int)rod);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by configureBOC: " << s.what() << endl;
    throw myRaise(s);
  }
}

Sct_SctApi::BOCMonitorSequence * SctApiServer::getBOCMonitorArray (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  try {
    const std::vector<double> workersBOCMonitorVec = worker().getBOCMonitorArray((unsigned int) partition, crate, rod);
    return copyVectorToCorba<Sct_SctApi::BOCMonitorSequence, double>(workersBOCMonitorVec);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by getBOCMonitorArray: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::flashLED (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::UShort slaveNumber, CORBA::ULong period, CORBA::ULong flashes)
{
  try {
    worker().flashLED((unsigned int)partition, (unsigned int)crate, (unsigned int)rod, slaveNumber, period, flashes);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by flashLED: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::echo (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, const Sct_SctApi::DataBlock & data)
{
  try {
    unsigned long ourLength = data.length();

    unsigned long *array = new unsigned long[ourLength];
    unsigned long *body = copyCorbaToArray(data, array);

    worker().echo((unsigned int)partition, (unsigned int)crate, (unsigned int)rod, (unsigned int)ourLength, body);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by echo: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::createDebugPrimList ()
{
  try {
    worker().createDebugPrimList();
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by createDebugPrimList: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::addDebugPrimList (CORBA::Long index, CORBA::Long id, CORBA::Long version, const Sct_SctApi::DataBlock & body)
{
  try {
    unsigned long ourLength = body.length();
    unsigned long *array = new unsigned long[ourLength];
    copyCorbaToArray(body, array);

    worker().addDebugPrimList(ourLength, index, id, version, array);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by addDebugPrimList: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::sendDebugPrimList (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  try {
    worker().sendDebugPrimList((unsigned int)partition, (unsigned int)crate, (unsigned int)rod);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by sendDebugPrimList: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::sendDebugSlavePrimList (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong slave, CORBA::Boolean await, CORBA::Boolean response)
{
  try {
    worker().sendDebugSlavePrimList((unsigned int)partition, (unsigned int)crate, (unsigned int)rod, slave, await, response);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by sendDebugSlavePrimList: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::setRunNumber (CORBA::ULong newRun)
{
  worker().setRunNumber(newRun);
}

void SctApiServer::setScanNumber (CORBA::ULong newScan)
{
  worker().setScanNumber(newScan);
}

CORBA::ULong SctApiServer::getRunNumber ()
{
  return worker().getRunNumber();
}

CORBA::ULong SctApiServer::getScanNumber ()
{
  return worker().getScanNumber();
}

void SctApiServer::unsetDebugOption (const char * opt)
{
  worker().unsetDebugOption(opt);
}

void SctApiServer::setDebugOption (const char * opt)
{
  worker().setDebugOption(opt);
}

Sct_SctApi::DebugOptionList * SctApiServer::listEnabledDebugOptions ()
{

  std::list<std::string> myList = worker().listEnabledDebugOptions();
  Sct_SctApi::DebugOptionList * result = SequenceUtils::newSafeSequence<Sct_SctApi::DebugOptionList>(myList.size());
  unsigned int i=0;
  for(std::list<std::string> :: const_iterator m = myList.begin();
      m!=myList.end();
      (++m,++i)) {
    (*result)[i] = CORBA::string_dup((*m).c_str());
  }

  return result;
}

Sct_SctApi::DebugOptionList * SctApiServer::listDebugOptions ()
{
  std::vector<std::string> myList = worker().listDebugOptions();

  Sct_SctApi::DebugOptionList * result = SequenceUtils::newSafeSequence<Sct_SctApi::DebugOptionList>(myList.size());
  unsigned int i=0;
  for(std::vector<std::string> :: const_iterator m = myList.begin();
      m!=myList.end();
      (++m,++i)) {
    (*result)[i] = CORBA::string_dup((*m).c_str());
  }

  return result;
}

void SctApiServer::standardRegisterDump (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  worker().standardRegisterDump(partition, crate, rod);
}

void SctApiServer::standardRegisterDumpAll ()
{
  worker().standardRegisterDumpAll();
}

sctConf::ModuleList * SctApiServer::getModuleList () 
{
  // LESTER result = _sctConf_T_ModuleList_sequence::Create(4, NULL);
  // LESTER result->Clear(0);

  std::list<std::string> myList = worker().getModuleList();
  sctConf::ModuleList * result = SequenceUtils::newSafeSequence<sctConf::ModuleList>(myList.size());

  unsigned int i=0;
  for(std::list<std::string> :: const_iterator m = myList.begin();
      m!=myList.end();
      (++m, ++i)) {
    (*result)[i] = CORBA::string_dup((*m).c_str());
  }

  return result;
}

CORBA::Boolean SctApiServer::isRODPresent (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  return worker().isRODPresent(partition, crate, rod);
}

void SctApiServer::loadConfiguration (const char * filename)
{
  try {
    worker().loadConfiguration(filename);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by loadConfiguration: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::storeModuleConfigurations ()
{
  try {
    worker().storeModuleConfigurations();
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by storeModuleConfigurations: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::configureAllModules ()
{
  try {
    worker().sendAllABCDModules(myTranslateBank(Sct_SctApi::PHYSICS_CONFIG));
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by configureAllModules: " << s.what() << endl;
    throw myRaise(s);
  }
}

sctConf::ABCDModule SctApiServer::retrieveModule (CORBA::ULong mid)
{
  ABCDModule *body = worker().retrieveModule(mid);

  return copyABCDModuleToCorba(*body);
}

CORBA::Short SctApiServer::dspBlockDump (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                                             CORBA::Long dspStart, CORBA::Long numWords, CORBA::Long dspNumber)
{
  return worker().dspBlockDump((unsigned int)partition, (unsigned int)crate, (unsigned int)rod,
                             dspStart, numWords, dspNumber);
}

Sct_SctApi::DataBlock * SctApiServer::dspBlockRead (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::Long dspStart, CORBA::Long numWords, CORBA::Long dspNumber)
{
  unsigned long length;
  unsigned long *body = worker().dspBlockRead((unsigned int)partition, (unsigned int)crate, (unsigned int)rod,
                                            dspStart, numWords, dspNumber, length);

  return copyArrayToCorba<Sct_SctApi::DataBlock, unsigned long>(body, length);
}

CORBA::Short SctApiServer::dspBlockWrite (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod,
                                              const Sct_SctApi::DataBlock & buffer, CORBA::ULong dspAddress, 
                                              CORBA::Long dspNumber)
{
  unsigned long length = buffer.length();
  unsigned long *array = new unsigned long[length];
  copyCorbaToArray(buffer, array);

  return worker().dspBlockWrite((unsigned int)partition, (unsigned int)crate, (unsigned int)rod, 
                                array, dspAddress, length, dspNumber);
}

CORBA::ULong SctApiServer::dspSingleRead (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                                          CORBA::ULong dspAddr, CORBA::Long dspNumber)
{
  return worker().dspSingleRead((unsigned int)partition, (unsigned int)crate, (unsigned int)rod, 
                              dspAddr, dspNumber);
}

void SctApiServer::dspSingleWrite (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong dspAddr, CORBA::ULong val, CORBA::Long dspNumber)
{
  worker().dspSingleWrite((unsigned int)partition, (unsigned int)crate, (unsigned int)rod, dspAddr, val, dspNumber);
}

Sct_SctApi::DataBlock * SctApiServer::readSlaveDsp (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::Short s, CORBA::ULong add, CORBA::ULong words)
{
  unsigned long *body = worker().primReadSlaveDsp((unsigned int)partition, (unsigned int)crate, (unsigned int)rod, s, add, words);

  return copyArrayToCorba<Sct_SctApi::DataBlock, unsigned long>(body, words);
}

CORBA::ULong SctApiServer::findModuleSerial (const char * sn)
{
  return worker().findModule(sn);
}

char * SctApiServer::convertToString (CORBA::ULong mid)
{
  return copyStringToCorba(worker().convertToString(mid));
}

CORBA::ULong SctApiServer::findModule (CORBA::UShort mur, CORBA::UShort number)
{
  return worker().findModule(mur, number);
}

CORBA::ULong SctApiServer::findModuleRod (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong channel) 
{
  UINT32 mid = 0;
  mid |= (partition & 0xff) << 24;
  mid |= (crate & 0xff) << 16;
  mid |= (rod & 0xff) << 8;
  mid |= (channel & 0xff) << 0;
  return mid;
}

CORBA::ULong SctApiServer::findBarrelModule (
                                             CORBA::ULong barrel, CORBA::ULong row, CORBA::Short number)
{
  return worker().findBarrelModule(barrel, row, number);
}

void SctApiServer::convertToBarrelModule (CORBA::ULong mid, CORBA::ULong & barrel, CORBA::ULong & row, CORBA::Short & number)
{

  UINT32 tmpBarrel = barrel;
  UINT32 tmpRow    = row   ;
  int    tmpNumber = number;

  worker().convertToBarrelModule(static_cast<unsigned int>(mid), tmpBarrel, tmpRow, tmpNumber);

  barrel = tmpBarrel;
  row    = tmpRow;
  number = tmpNumber;
}

CORBA::ULong SctApiServer::findEndcapModule (CORBA::Long disk, CORBA::Long ring, CORBA::Long number)
{
  return worker().findEndcapModule(disk, ring, number);
}

void SctApiServer::convertToEndcapModule (CORBA::ULong mid, CORBA::Long & disk, CORBA::ULong & ring, CORBA::ULong & number)
{

   INT32 tmpDisk   = disk;
  UINT32 tmpRing   = ring;
  UINT32 tmpNumber = number;

  worker().convertToEndcapModule((unsigned int)mid, tmpDisk, tmpRing, tmpNumber);

  disk   = tmpDisk;
  ring   = tmpRing;
  number = tmpNumber;
}

void SctApiServer::getABCDModule (CORBA::ULong mid, Sct_SctApi::BankType bank)
{
  try {
    worker().getABCDModule(mid, myTranslateBank(bank));
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by getABCDModule: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::getABCDModules (Sct_SctApi::BankType bank)
{
  try {
    worker().getABCDModules(myTranslateBank(bank));
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by getABCDModules: " << s.what() << endl;
    throw myRaise(s);
  }
}

sctConf::ABCDModule SctApiServer::getABCDModuleRaw (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                                                       CORBA::ULong slot, Sct_SctApi::BankType bank)
{
  try {
    boost::shared_ptr<ABCDModule> body = worker().getABCDModuleRaw(partition, crate, rod, slot, myTranslateBank(bank));
    return copyABCDModuleToCorba(*(body.get()));
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by getABCDModules: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::setABCDModule (CORBA::ULong mid, Sct_SctApi::BankType bank)
{
  try {
    worker().setABCDModule(mid, myTranslateBank(bank));
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by setABCDModule: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::setABCDModules (Sct_SctApi::BankType bank)
{
  try {
    worker().setABCDModules(myTranslateBank(bank));
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by setABCDModules: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::sendABCDModule (CORBA::ULong mid, Sct_SctApi::BankType bank, CORBA::ULong type)
{
  try {
    worker().sendABCDModule(mid, myTranslateBank(bank), SctApi::ConfigType(type));
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by sendABCDModule: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::sendABCDModules (Sct_SctApi::BankType bank)
{
  try {
    worker().sendAllABCDModules(myTranslateBank(bank));
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by sendABCDModules: " << s.what() << endl;
    throw myRaise(s);
  }
}

void SctApiServer::modifyABCDMask (CORBA::ULong mid, const Sct_SctApi::DataBlock & mask)
{
  UINT32 array[48];
  worker().modifyABCDMask(mid, copyCorbaToArray(mask, array));
}

void SctApiServer::modifyABCDTrims (CORBA::ULong mid, const Sct_SctApi::CharBlock & trims){
  UINT8 array[12*128];
  worker().modifyABCDTrims (mid, copyCorbaToArray(trims, array) );
} 

void SctApiServer::modifyABCDRC (CORBA::ULong mid, CORBA::ULong chip, CORBA::UShort index, CORBA::Double p0, CORBA::Double p1, CORBA::Double p2)
{
  cout << "SctApiServer::modifyABCDRC called" << endl;
  worker().modifyABCDRC(mid, chip, index, p0, p1, p2);
}

void SctApiServer::modifyABCDVar (CORBA::ULong typ, CORBA::Double var)
{
  worker().modifyABCDVar(typ, var);
}

void SctApiServer::modifyABCDModuleVar (CORBA::ULong mid, CORBA::ULong typ, CORBA::Double var)
{
  worker().modifyABCDVar(mid, typ, var);
}

void SctApiServer::modifyABCDChipVar (CORBA::ULong mid, CORBA::ULong chip, CORBA::ULong typ, CORBA::Double var)
{
  worker().modifyABCDVar(mid, chip, typ, var);
}

void SctApiServer::modifyABCDVarROD (CORBA::ULong typ, CORBA::Double var, Sct_SctApi::BankType bank)
{
  worker().modifyABCDVarROD(typ, var, myTranslateBank(bank));
}

void SctApiServer::modifyABCDModuleVarROD (CORBA::ULong mid, CORBA::ULong typ, CORBA::Double var, Sct_SctApi::BankType bank)
{
  worker().modifyABCDVarROD(mid, typ, var, myTranslateBank(bank));
}

void SctApiServer::modifyBOCParam (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                                   CORBA::ULong channel, CORBA::ULong type, CORBA::ULong val)
{
  worker().modifyBOCParam(partition, crate, rod, channel, type, val);
}

void SctApiServer::lasersOff ()
{
  worker().lasersOff();
}

void SctApiServer::timSetFrequency (CORBA::ULong partition, CORBA::ULong crate, 
                                    CORBA::Double trigFreq, CORBA::Double rstFreq)
{
  worker().timSetFrequency(partition, crate, trigFreq, rstFreq);
}

void SctApiServer::freeTriggers (CORBA::ULong partition, CORBA::ULong crate)
{
  worker().freeTriggers(partition, crate);
}

void SctApiServer::stopTriggers (CORBA::ULong partition, CORBA::ULong crate)
{
  worker().stopTriggers(partition, crate);
}

void SctApiServer::timL1A (CORBA::ULong partition, CORBA::ULong crate)
{
  worker().timL1A(partition, crate);
}

void SctApiServer::timCalL1A (CORBA::ULong partition, CORBA::ULong crate, CORBA::Short delay)
{
  worker().timCalL1A(partition, crate, delay);
}

CORBA::Short SctApiServer::timReadRegister(CORBA::ULong partition,
					   CORBA::ULong crate,
					   CORBA::Short value) 
{
  return worker().timReadRegister(partition, crate, value);
}

void SctApiServer::timWriteRegister (CORBA::ULong partition, CORBA::ULong crate, CORBA::Short reg, CORBA::Short value)
{
  worker().timWriteRegister(partition, crate, reg, value);
}

void SctApiServer::sendTimTrigger (CORBA::ULong partition,
				   CORBA::ULong crate,
				   Sct_SctApi::Trigger_ptr trig) 
{
  const TriggerWrapper trigWrapper(trig, TriggerWrapper::doNotTakeOwnershipOfObjectReference);
  worker().sendTimTrigger(partition, crate, &trigWrapper);
}

void SctApiServer::sendTrigger (CORBA::ULong partition,
				CORBA::ULong crate,
				CORBA::ULong rod,
				Sct_SctApi::Trigger_ptr trig)
{
  const TriggerWrapper trigWrapper(trig, TriggerWrapper::doNotTakeOwnershipOfObjectReference);
  worker().sendTrigger(partition, crate, rod, &trigWrapper);
}

void SctApiServer::doScan(Sct_SctApi::Scan_ptr thisScan)
{
  mrsStream << "SCTAPI_SCAN" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("Starting scan") << ENDM;

  ISInfoInt scanning(1);
  isDict.insert("SCTAPIServer.scanning", scanning);

  //ScanServer *serverScan = dynamic_cast<ScanServer*>(thisScan);
  //IPCObjectVar<ScanServer>::duplicate(serverScan);
  
  const boost::shared_ptr< ::SctApi::Scan> scanWrapper 
    = boost::shared_ptr< ::SctApi::Scan>(new ScanWrapper(thisScan/*, ScanWrapper::doNotTakeOwnershipOfObjectReference*/));

  try {
    worker().doScan(scanWrapper);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by doScan: " << s.what() << endl;

//     serverScan->decRef();
    isDict.remove("SCTAPIServer.scanning");

    mrsStream << "SCAN_ABORT" << MRS_ERROR << MRS_QUALIF("SctApiServer") << MRS_TEXT((string("Scan aborted: ") + s.what()).c_str()) << ENDM;
    throw myRaise(s);
  }

  // Don't decrease reference, so polling thread can keep it
//   realScan->decRef();
}

void SctApiServer::doRawScan (Sct_SctApi::Scan_ptr thisScan,
			      CORBA::Long delay,
			      CORBA::Long width,
			      CORBA::Boolean configureModules,
			      CORBA::Boolean clockByTwo)
{
  mrsStream << "SCTAPI_RAWSCAN" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("Raw scan started") << ENDM;

  ISInfoInt scanning(1);
  isDict.insert("SCTAPIServer.scanning", scanning);

  // ScanServer *realScan = dynamic_cast<ScanServer*>(thisScan);
  //IPCObjectVar<ScanServer>::duplicate(realScan);

  boost::shared_ptr< ::SctApi::Scan> scanWrapper
    = boost::shared_ptr< ::SctApi::Scan>(new ScanWrapper(thisScan/*, ScanWrapper::doNotTakeOwnershipOfObjectReference*/));

  try {
    worker().doRawScan(scanWrapper, delay, width, configureModules, clockByTwo);
  } catch(SctApi::SctApiException &s) {
    cout << "Exception thrown by doRawScan: " << s.what() << endl;

    isDict.remove("SCTAPIServer.scanning");

    mrsStream << "SCAN_RAW_ABORT" << MRS_ERROR << MRS_QUALIF("SctApiServer") << MRS_TEXT((string("Raw scan aborted: ") + s.what()).c_str()) << ENDM;
    throw myRaise(s);
  }

  isDict.remove("SCTAPIServer.scanning");

  mrsStream << "SCTAPI_RAWSCAN" << MRS_INFORMATION << MRS_QUALIF("SctApiServer") << MRS_TEXT("Raw scan completed") << ENDM;
}

void SctApiServer::printBOCSetup (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  worker().printBOCSetup(partition, crate, rod);
}

Sct_SctApi::BOCChannelList * SctApiServer::currentBOCSetup (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod) 
{
  return copyBOCChannelConfigVectorToCorba(worker().currentBOCSetup(partition, crate, rod));
}

sctConf::BOCGlobalConfig SctApiServer::currentBOCRegisters (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod) 
{
  return copyBOCGlobalConfigToCorba(worker().currentBOCRegisters(partition, crate, rod));
}

void SctApiServer::printABCDModule (CORBA::ULong mid)
{
  worker().printABCDModule(mid);
}

void SctApiServer::printABCDRodModule (CORBA::ULong mid, Sct_SctApi::BankType bank)
{
  worker().printABCDRodModule(mid, myTranslateBank(bank));
}

void SctApiServer::decodeEvent (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::Short sl, CORBA::Short index, CORBA::Boolean extFlag, CORBA::Boolean errorType)
{
  worker().decodeEvent(partition, crate, rod, sl, index, extFlag, errorType);
}

void SctApiServer::rawData (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::Long delay, CORBA::Long units, CORBA::Boolean setMask)
{
  worker().rawData(partition, crate, rod, delay, units, setMask);
}

void SctApiServer::rawDataWithTrigger (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::Long delay, CORBA::Long units, CORBA::Boolean setMask, Sct_SctApi::Trigger_ptr trig)
{
  const TriggerWrapper trigWrapper(trig, TriggerWrapper::doNotTakeOwnershipOfObjectReference);
  worker().rawData(partition, crate, rod, delay, units, setMask, &trigWrapper);
}

void SctApiServer::sendL1A (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  worker().sendL1A(partition, crate, rod);
}

Sct_SctApi::CharBlock * SctApiServer::probe (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  return copyVectorToCorba<Sct_SctApi::CharBlock, char>(worker().probe(partition, crate, rod));
}

Sct_SctApi::CharBlock * SctApiServer::probeHarness (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong harness)
{
  return copyVectorToCorba<Sct_SctApi::CharBlock, char>(worker().probe(partition, crate, rod, harness));
}

Sct_SctApi::CharBlock * SctApiServer::probeWithTrigger (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, Sct_SctApi::Trigger_ptr trig)
{
  const TriggerWrapper trigWrapper(trig, TriggerWrapper::doNotTakeOwnershipOfObjectReference);
  vector<char> ret = worker().probeWithTrigger(partition, crate, rod, &trigWrapper);
  return copyVectorToCorba<Sct_SctApi::CharBlock, char>(ret);
}

Sct_SctApi::CharBlock * SctApiServer::probeHarnessWithTrigger (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, Sct_SctApi::Trigger_ptr trig, CORBA::ULong harness)
{
  const TriggerWrapper trigWrapper(trig, TriggerWrapper::doNotTakeOwnershipOfObjectReference);
  vector<char> ret = worker().probeWithTrigger(partition, crate, rod, &trigWrapper, harness);
  return copyVectorToCorba<Sct_SctApi::CharBlock, char>(ret);
}

Sct_SctApi::CharMatrix * SctApiServer::probeScanHarnessWithTrigger (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, Sct_SctApi::Scan_ptr scan, CORBA::ULong harness)
{
  const boost::shared_ptr< ::SctApi::Scan> scanWrapper 
    = boost::shared_ptr< ::SctApi::Scan>( new ScanWrapper(scan/*, ScanWrapper::doNotTakeOwnershipOfObjectReference*/));
  vector<vector<char> > ret = worker().probeScan(partition, crate, rod, scanWrapper, harness);
  return copyVectorVectorToCorba<Sct_SctApi::CharMatrix, Sct_SctApi::CharBlock, char>(ret);
}

CORBA::Boolean SctApiServer::checkAllModulesProbe (const char * values)
{
  return worker().checkAllModulesProbe(values);
}

Sct_SctApi::AutoResultList * SctApiServer::autoConfigureScan ()
{
  ::SctApi::AutoConf::AutoConfigurer configurer(worker());
  configurer.run();

  return copyAutoResultListToCorba(configurer.getResults()); // worker().autoConfigureScan();
}

CORBA::ULong SctApiServer::getRodSlot (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  return worker().getRodSlot(partition, crate, rod);
}

void SctApiServer::testLinkOutSelect (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong link)
{
  worker().testLinkOutSelect(partition, crate, rod, link);
}

CORBA::ULong SctApiServer::testLinkOutRetrieve (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)
{
  return worker().testLinkOutRetrieve(partition, crate, rod);
}

void SctApiServer::requestHardReset (CORBA::ULong mid)
{
  worker().requestHardReset(mid);
}

void SctApiServer::requestIVCurve (CORBA::ULong mid, CORBA::Float start, CORBA::Float stop, CORBA::Float step, CORBA::UShort delay, CORBA::Float currentLim) 
{
  worker().requestIVCurve(mid, start, stop, step, delay, currentLim);
}

void SctApiServer::setSelect (CORBA::ULong mid, CORBA::Boolean value)
{
  worker().setSelect(mid, value);
}

void SctApiServer::setSelectAllInCrate (CORBA::Long crate, CORBA::Boolean value)
{
  worker().setSelectAllInCrate(crate, (bool)value, ddcMultipleTimeout);
}

void SctApiServer::changeRunState (CORBA::Long state)
{
  worker().changeRunState(state);
}

void SctApiServer::ddcSetSingleTimeout (CORBA::Short timeout)
{
  ddcSingleTimeout = timeout;
}

CORBA::Short SctApiServer::ddcGetSingleTimeout () 
{
  return ddcSingleTimeout;
}

void SctApiServer::ddcSetMultipleTimeout (CORBA::Short timeout)
{
  ddcMultipleTimeout = timeout;
}

CORBA::Short SctApiServer::ddcGetMultipleTimeout ()
{
  return ddcMultipleTimeout;
}

void SctApiServer::ddcChangeState (CORBA::Long crate, CORBA::Long channel, CORBA::Short state)
{
  worker().ddcChangeState(crate, channel, state, ddcMultipleTimeout);
}

void SctApiServer::ddcChangeStateAll (CORBA::Long crate, CORBA::Short state)
{
  worker().ddcChangeStateAll(crate, state, ddcMultipleTimeout);
}

void SctApiServer::ddcLoadConfiguration (CORBA::Long crate, CORBA::Short state)
{
  worker().ddcLoadConfiguration(crate, state, ddcMultipleTimeout);
}

void SctApiServer::ddcHardReset (CORBA::Long crate, CORBA::Long channel)
{
  worker().ddcHardReset(crate, channel, ddcSingleTimeout);
}

void SctApiServer::ddcHardResetAll (CORBA::Long crate)
{
  worker().ddcHardResetAll(crate, ddcMultipleTimeout);
}

CORBA::ULong SctApiServer::ddcGetChannelParameterInt (CORBA::Long crate, CORBA::Long channel, const char * name)
{
  return worker().ddcGetChannelParameterInt(crate, channel, string(name), ddcSingleTimeout);
}

CORBA::Float SctApiServer::ddcGetChannelParameterFloat (CORBA::Long crate, CORBA::Long channel, const char * name)
{
  return worker().ddcGetChannelParameterFloat(crate, channel, string(name), ddcSingleTimeout);
}

char * SctApiServer::ddcGetChannelParameterString (CORBA::Long crate, CORBA::Long channel, const char * name)
{
  return copyStringToCorba(worker().ddcGetChannelParameterString(crate, channel, string(name), ddcSingleTimeout));
}

CORBA::ULong SctApiServer::ddcGetCardParameterInt (CORBA::Long crate, CORBA::Long card, const char * name)
{
  return worker().ddcGetCardParameterInt(crate, card, string(name), ddcSingleTimeout);
}

CORBA::Float SctApiServer::ddcGetCardParameterFloat (CORBA::Long crate, CORBA::Long card, const char * name)
{
  return worker().ddcGetCardParameterFloat(crate, card, string(name), ddcSingleTimeout);
}

char * SctApiServer::ddcGetCardParameterString (CORBA::Long crate, CORBA::Long card, const char * name)
{
  return copyStringToCorba(worker().ddcGetCardParameterString(crate, card, string(name), ddcSingleTimeout));
}

CORBA::ULong SctApiServer::ddcGetCrateParameterInt (CORBA::Long crate, const char * name)
{
  return worker().ddcGetCrateParameterInt(crate, string(name), ddcSingleTimeout);
}

CORBA::Float SctApiServer::ddcGetCrateParameterFloat (CORBA::Long crate, const char * name)
{
  return worker().ddcGetCrateParameterFloat(crate, string(name), ddcSingleTimeout);
}

char * SctApiServer::ddcGetCrateParameterString (CORBA::Long crate, const char * name)
{
  return copyStringToCorba(worker().ddcGetCrateParameterString(crate, string(name), ddcSingleTimeout));
}

void SctApiServer::ddcGetCrateParameters (CORBA::Long crate, const Sct_SctApi::NameList & names, Sct_SctApi::NameValueList_out nameValues) 
{
  const std::vector<std::string> & myNames = copyCorbaToVector<Sct_SctApi::NameList, std::string>(names);

  std::map<std::string, float> myNameValues;

  worker().ddcGetCrateParameters(crate, myNames, myNameValues, ddcMultipleTimeout);

  nameValues = copyMapToNameValueList(myNameValues);
}

Sct_SctApi::Scan_ptr SctApiServer::createScan ()
{
  std::cout << "Asked to create a scan\n";

  SharedScanServer defScan = ::Sct::DestroyingDeleter<ScanServer>::shared_ptr( new ScanServer ); // , *defaultScan);

  scans.push_back(defScan);

  return defScan->_this();
}

Sct_SctApi::Scan_ptr SctApiServer::getDefaultScan ()
{
  std::cout << "Asked for a reference to the default scan\n";

  return defaultScan->_this();
}

Sct_SctApi::Trigger_ptr SctApiServer::createTrigger ()
{
  std::cout << "Asked to create a trigger\n";

  TriggerServer * defTrig = new TriggerServer;
#warning DOES ANYONE EVER DELETE THIS? I THINK NOT! (ChrisLester)
  
  return defTrig->_this();
}

Sct_SctApi::Trigger_ptr SctApiServer::getDefaultTrigger ()
{
  std::cout << "Asked for the default trigger\n";
#warning I DONT THINK ANYONE ACTUALLY SETS THE DEFAULT TRIGGER! THIS IS AN INVITATION TO SEG-FAULT! (ChrisLester)
  std::cout << "I will now do my best to segmentation fault." << std::endl;
  assert(defaultTrigger);
  return defaultTrigger->_this();
}

Sct_SctApi::ScanList * SctApiServer::knownScans ()
{
  std::cout << "In known scans\n";

  Sct_SctApi::ScanList * myScans = SequenceUtils::newSafeSequence<Sct_SctApi::ScanList>(scans.size());

  std::cout << "Copying sequence of length " << scans.size() << std::endl;

  unsigned int i=0;
  for(Scans::iterator it=scans.begin();
      it != scans.end();
      (++i,++it)) {
    ScanServer & thisScan = **it;
    (*myScans)[i] = thisScan._this();
  }

  return myScans;
}

SctApi::SctApi &SctApiServer::worker() 
{
  return api;
}

/**************** ScanServer ****************/

ScanServer::ScanServer() 
  : IPCNamedObject<POA_Sct_SctApi::Scan>(Sct::SctNames::getPartition(),
					 getServerName().c_str()), 
    m_worker(new SctApi::ScanDefImpl())
{
  std::cout << "Create a new ScanServer\n";

//    dynamic_cast<::Scan*>(this)->print();
//   std::cout << "trigsPerBurst = " << worker().getNTrigs() << std::endl; 

  boost::shared_ptr< ::SctApi::Trigger> trig1 
    = boost::shared_ptr< ::SctApi::Trigger>(new ::SctApi::TriggerImpl);
  boost::shared_ptr< ::SctApi::Trigger> trig2 
    = boost::shared_ptr< ::SctApi::Trigger>(new ::SctApi::TriggerImpl);

  std::cout << "TrigServer : " ;
  worker().setTrigger1( trig1 );
  worker().setTrigger2( trig2 );
  std::cout << " getTrigger1() " << worker().getTrigger1();
  std::cout << " getTrigger2() " << worker().getTrigger2();
  std::cout << std::endl;

//    if(!publish()) {
//      cerr << "Failed to publish new scan" << std::endl;
//    }

//    std::cout << "Published scan\n";
  std::cout << " ... ScanServer created\n";
}

ScanServer::ScanServer(const ScanServer &copy) 
  : IPCNamedObject<POA_Sct_SctApi::Scan>(Sct::SctNames::getPartition(),
					 getServerName().c_str()),
    m_worker(SctApi::ScanDefImpl::clone(copy.m_worker))
{
  std::cout << "Create a copy of a ScanServer\n";

//   TriggerServer *temp = new TriggerServer(ipcServer.get());
//   worker().setTrigger1(IPCObjectVar<TriggerServer>::duplicate(temp)->m_worker);

//   TriggerServer *temp2 = new TriggerServer(ipcServer.get());
//   worker().setTrigger2(IPCObjectVar<TriggerServer>::duplicate(temp2)->m_worker);

#warning "Need getTrigger to create new TriggerServer if none found?"

  std::cout << " ... ScanServer copied\n";
}

ScanServer::~ScanServer() 
{
//    return IPCObjectVar<ScanServer>::release(this);
  
//    withdraw();
}

void ScanServer::shutdown() 
{
  SctApiServer::Scans &scans = SctApiServer::singleton->scans;

  for (SctApiServer::Scans::iterator item = scans.begin();
       item != scans.end();
       ++item) {
    if ((*item).get() == this) {
      scans.erase(item);
      return;
    };
  };
  std::cout << "*******Couldn't find destroyed scan in scans list**********!\n";
}


int ScanServer::nextNumber = 0;

std::string ScanServer::getServerName() {
  std::ostringstream serverName;
  serverName << "ScanServer" << nextNumber++;
  return serverName.str();
}

void ScanServer::print() 
{
  std::cout << "Try to print a scan\n";
  worker().print();
  std::cout << "Gave up or sucess (printing scan)\n";
}

void ScanServer::configure (CORBA::UShort type, CORBA::Double start, CORBA::Double stop, CORBA::Double step) 
{
  worker().configure(type, start, stop, step);
}

void ScanServer::configure2 (CORBA::UShort type, CORBA::Double start, CORBA::Double stop, CORBA::Double step) 
{
  worker().configure2(type, start, stop, step);
}

Sct_SctApi::Trigger_ptr ScanServer::getTrigger1 () 
{
  const boost::shared_ptr< ::SctApi::Trigger> t = worker().getTrigger1();
  TriggerServer * ts = new TriggerServer(t);
  // Now at this point one might worry that ts might never be deleted, but, the intention is that since TriggerServer inherits from PortableServer::RefCountServantBase then as soon as the last ref to it is removed, then it should be deleted.  Touch wood.  So we had better make the first ref:
  return ts->_this();
}

Sct_SctApi::Trigger_ptr ScanServer::getTrigger2 () 
{
  const boost::shared_ptr< ::SctApi::Trigger> t = worker().getTrigger2();
  TriggerServer * ts = new TriggerServer(t);
  // Now at this point one might worry that ts might never be deleted, but, the intention is that since TriggerServer inherits from PortableServer::RefCountServantBase then as soon as the last ref to it is removed, then it should be deleted.  Touch wood.  So we had better make the first ref:
  return ts->_this();
}

void ScanServer::setTrigger1 (Sct_SctApi::Trigger_ptr trig)
{
  //  TriggerServer *server = dynamic_cast<TriggerServer *>(trig);
  boost::shared_ptr< ::SctApi::Trigger> trigWrapper 
    = boost::shared_ptr< ::SctApi::Trigger>(new TriggerWrapper(trig, TriggerWrapper::doNotTakeOwnershipOfObjectReference));
  worker().setTrigger1(trigWrapper);
}

void ScanServer::setTrigger2 (Sct_SctApi::Trigger_ptr trig) 
{
  //  TriggerServer *server = dynamic_cast<TriggerServer *>(trig);
  boost::shared_ptr< ::SctApi::Trigger> trigWrapper 
    = boost::shared_ptr< ::SctApi::Trigger>(new TriggerWrapper(trig, TriggerWrapper::doNotTakeOwnershipOfObjectReference));
  worker().setTrigger2(trigWrapper);
}

void ScanServer::setScanVariable1 (CORBA::UShort var) 
{
  worker().setScanVariable1(var);
}

void ScanServer::setScanVariable2 (CORBA::UShort var) 
{
  worker().setScanVariable2(var);
}

CORBA::UShort ScanServer::getScanVariable1 () 
{
  return worker().getScanVariable1();
}

CORBA::UShort ScanServer::getScanVariable2 () 
{
  return worker().getScanVariable2();
}

void ScanServer::setNTrigs (CORBA::ULong nTrigs) 
{
  worker().setNTrigs(nTrigs);
}

CORBA::ULong ScanServer::getNTrigs () 
{
  return worker().getNTrigs();
}

/*
// Hopefully this function is now obsolete.
static void enumerateToLongList(CORBA::ULong val, void *list) {
SctApi::Scan::TrigPoints &myList = *(SctApi::Scan::TrigPoints *)list;

myList.push_back(val);
}
*/

Sct_SctApi::Scan::TrigPoints * ScanServer::getVariableTrigs() {
  Sct_SctApi::Scan::TrigPoints * result = SequenceUtils::newSafeSequence<Sct_SctApi::Scan::TrigPoints>(worker().getVariableTrigs().size());

  for(unsigned long i=0; i<worker().getVariableTrigs().size(); ++i) {
    (*result)[i] = worker().getVariableTrigs()[i];
  }

  return result;
};

void ScanServer::setVariableTrigs(const Sct_SctApi::Scan::TrigPoints& scans) {
  SctApi::Scan::TrigPoints newPoints(scans.length());
  for (unsigned int i=0; i<scans.length(); ++i) {
    newPoints[i] = scans[i];
  };
  worker().setVariableTrigs(newPoints);
};

void ScanServer::setVariableTrigRange(CORBA::UShort start, CORBA::UShort end, CORBA::ULong value) {
  worker().setVariableTrigRange(start, end, value);
};

Sct_SctApi::Scan::ScanPoints * ScanServer::getScanPoints1 () 
{
  Sct_SctApi::Scan::ScanPoints * result = SequenceUtils::newSafeSequence<Sct_SctApi::Scan::ScanPoints>(worker().getScanPoints1().size());

  for(unsigned long i=0; i<worker().getScanPoints1().size(); ++i) {
    (*result)[i] = worker().getScanPoints1()[i];
  }

  return result;
}

Sct_SctApi::Scan::ScanPoints * ScanServer::getScanPoints2 ()
{
  Sct_SctApi::Scan::ScanPoints * result = SequenceUtils::newSafeSequence<Sct_SctApi::Scan::ScanPoints>(worker().getScanPoints2().size());

  for(unsigned long i=0; i<worker().getScanPoints2().size(); ++i) {
    (*result)[i] = worker().getScanPoints2()[i];
  }

  return result;
}

/*
// Hopefully this function is now obsolete.
static void enumerateToRealList(CORBA::Double val, void *list) {
SctApi::Scan::ScanPoints &myList = *(SctApi::Scan::ScanPoints *)list;

myList.push_back(val);
}
*/

void ScanServer::setScanPoints1 (const Sct_SctApi::Scan::ScanPoints & scans) 
{
  SctApi::Scan::ScanPoints newPoints(scans.length());
  for (unsigned int i=0; i<scans.length(); ++i) {
    newPoints[i] = scans[i];
  };
  //scans.Enumerate(enumerateToRealList, &newPoints);
  worker().setScanPoints1(newPoints);
}
 
void ScanServer::setScanPoints2 (const Sct_SctApi::Scan::ScanPoints & scans) 
{
  SctApi::Scan::ScanPoints newPoints(scans.length());
  for (unsigned int i=0; i<scans.length(); ++i) {
    newPoints[i] = scans[i];
  };
  //scans.Enumerate(enumerateToRealList, &newPoints);
  worker().setScanPoints2(newPoints);
}

CORBA::Long ScanServer::getOption (Sct_SctApi::Scan::ScanOptions opt) 
{
  return worker().getOption((SctApi::Scan::ScanOptions)opt);
}

void ScanServer::setOption (Sct_SctApi::Scan::ScanOptions opt, CORBA::Long val) 
{
  worker().setOption((SctApi::Scan::ScanOptions)opt, val);
}

void ScanServer::setScanNumber (CORBA::UShort scan) {
  worker().setScanNumber(scan);
};

void ScanServer::setRunNumber (CORBA::UShort scan) {
  worker().setRunNumber(scan);
};


CORBA::ULong ScanServer::getScanNumber () 
{
  return worker().getScanNumber();
}
 
CORBA::ULong ScanServer::getRunNumber () 
{
  return worker().getRunNumber();
}


sctConf::ModuleList * ScanServer::getModuleList (CORBA::ULong group) 
{

  std::list<std::string> myList = worker().getModuleList(group);
  sctConf::ModuleList * result = SequenceUtils::newSafeSequence<sctConf::ModuleList>(myList.size());
  unsigned int i=0;
  for(std::list<std::string> :: const_iterator m = myList.begin();
      m!=myList.end();
      (++i,++m)) {
    (*result)[i] = CORBA::string_dup((*m).c_str());
  }

  return result;
}

void ScanServer::setModuleList (CORBA::ULong group, const sctConf::ModuleList& newList) {

  ::SctApi::ModuleList tmp;
  const unsigned int siz = newList.length();

  for (unsigned int i=0; i<siz; ++i) {
    tmp.push_back(   (newList[i])._NP_ref()   );
  };

  worker().setModuleList( group, tmp );

};




CORBA::ULong ScanServer::getNGroups () 
{
  return worker().getNGroups();
}
char * ScanServer::getStartTime()
{
  string startTime =  boost::posix_time::to_iso_string(worker().getStartTime()).substr(0, 15); 
  return copyStringToCorba(startTime);
}
char * ScanServer::getEndTime()
{
  string endTime = boost::posix_time::to_iso_string(worker().getEndTime()).substr(0, 15); 
  return copyStringToCorba(endTime);    
}

void ScanServer::setNGroups(CORBA::UShort val) {
  worker().setNGroups(val);
}
void ScanServer::setStartTime(const char* t) {
  const boost::posix_time::ptime pt =  boost::posix_time::from_iso_string(t);
  worker().setStartTime(pt);
}
void ScanServer::setEndTime(const char* t) {
  const boost::posix_time::ptime pt =  boost::posix_time::from_iso_string(t);
  worker().setEndTime(pt);
}


SctApi::Scan &ScanServer::worker() 
{
  return *m_worker;
}



/******************* TriggerServer **********************/

TriggerServer::TriggerServer() 
  : IPCNamedObject<POA_Sct_SctApi::Trigger>(Sct::SctNames::getPartition(),
					    getServerName().c_str()),
    m_worker(new SctApi::TriggerImpl()) {
}

TriggerServer::TriggerServer(const boost::shared_ptr< ::SctApi::Trigger> other) 
  : IPCNamedObject<POA_Sct_SctApi::Trigger>(Sct::SctNames::getPartition(),
					    getServerName().c_str()),
    m_worker(other) {
}

TriggerServer::~TriggerServer()
{
}

void TriggerServer::shutdown() 
{
#warning THIS IS THE OLD destroy(ipcStatus *_status) METHOD, AND NOW IT IS THE SHUTDOWN METHOD.  THIS IS QUITE PROBABLY THE WRONG THING TO DO WITH THIS METHOD ... I DONT KNOW!
#warning PARTICULARLY BEWARE THE COMMENTING OF decRef();
  // decRef();
}

/*
 // Hopefully this member is now obsolete!
std::map<boost::shared_ptr< ::SctApi::Trigger>, TriggerServer *> TriggerServer::serverMap;
*/

/*
 // Hopefully this method is now obsolete!
TriggerServer *TriggerServer::getServer(boost::shared_ptr< ::SctApi::Trigger> trigger) {
  cout << "Looking for server for " << trigger.get() << endl;

  TriggerServer *result = serverMap[trigger];
  cout << " ... Found " << result << endl;

  return result;
  //  return dynamic_cast<TriggerServer *>(trigger.get());
}
*/

/*
 // Hopefully this method is now obsolete!
void TriggerServer::mapServer(boost::shared_ptr< ::SctApi::Trigger> trigger, TriggerServer *server) {
  cout << "Mapping trigger " << trigger.get() << " to  " << server << endl;
  server->incRef();
  serverMap.insert(make_pair(trigger, server));
}
*/

int TriggerServer::nextNumber = 0;

std::string TriggerServer::getServerName() {
  std::ostringstream serverName;
  serverName << "TriggerServer" << nextNumber++;
  std::cout << "Trigger name: " << serverName.str() << std::endl;;
  return serverName.str();
}

void TriggerServer::print () 
{
  std::cout << "Try to print a trigger\n";
  worker()->print();
  std::cout << "Succeeded in printing a trigger\n";
}

void TriggerServer::singleL1A () 
{
  worker()->singleL1A();
}

void TriggerServer::doubleL1A (CORBA::UShort delay) 
{
  worker()->doubleL1A(delay);
}

void TriggerServer::delayedL1A (CORBA::UShort delay) 
{
  worker()->delayedL1A(delay);
}

void TriggerServer::calL1A (CORBA::UShort delay) 
{
  worker()->calL1A(delay);
}

void TriggerServer::pulseL1A (CORBA::UShort delay) 
{
  worker()->pulseL1A(delay);
}

void TriggerServer::softL1A (CORBA::UShort delay) 
{
  worker()->softL1A(delay);
}

void TriggerServer::softCalL1A (CORBA::UShort delay, CORBA::UShort delay2) 
{
  worker()->softCalL1A(delay, delay2);
}

void TriggerServer::softPulseL1A (CORBA::UShort delay, CORBA::UShort delay2) 
{
  worker()->softPulseL1A(delay, delay2);
}

void TriggerServer::softBc (CORBA::UShort delay) 
{
  worker()->softBc(delay);
}

void TriggerServer::bcL1A (CORBA::UShort delay) 
{
  worker()->bcL1A(delay);
}

void TriggerServer::bcCalL1A (CORBA::UShort delay, CORBA::UShort delay2) 
{
  worker()->bcCalL1A(delay, delay2);
}

void TriggerServer::bcPulseL1A (CORBA::UShort delay, CORBA::UShort delay2) 
{
  worker()->bcPulseL1A(delay, delay2);
}

void TriggerServer::setCommIncr (CORBA::UShort command, CORBA::UShort incr)
{
  worker()->setCommIncr(command, incr);
}

void TriggerServer::getCommIncr (CORBA::UShort & command, CORBA::UShort & incr)
{
  worker()->getCommIncr(command, incr);
}

Sct_SctApi::RODTriggers * TriggerServer::getRODTriggers() {

  const ::SctApi::Trigger::RODTriggers & rts = worker()->getRODTriggers();
  Sct_SctApi::RODTriggers * result = SequenceUtils::newSafeSequence<Sct_SctApi::RODTriggers>(rts.size());

  for (unsigned int i=0; i<rts.size(); ++i) {
    const ::SctApi::Trigger::RODTrigger & rti = rts[i];

    Sct_SctApi::RODTrigger temp;
    temp.first  = rti.first;
    temp.second = rti.second;

    (*result)[i] = temp;
  };
  
  return result;
}

SctApi::BankType myTranslateBank(Sct_SctApi::BankType bank) {
  SctApi::BankType result;
  switch(bank) {
  case Sct_SctApi::PHYSICS_CONFIG:
    result = ::SctApi::SCTAPI_BANK_PHYSICS;
    break;
  case Sct_SctApi::CALIBRATION_CONFIG:
    result = SctApi::SCTAPI_BANK_CALIBRATION;
    break;
  case Sct_SctApi::SCAN_CONFIG:
    result = SctApi::SCTAPI_BANK_SCAN;
    break;
  default:
    result = SctApi::SCTAPI_BANK_PHYSICS;
    break;
  }
  return result;
}

//typedef Sct_SctApi_T_ilu__prefix_idlExceptionType_SctApiException iluSctApiException;

Sct_SctApi::SctApiException myRaise( ::SctApi::SctApiException &c) 
{
  cout << "Raising exception " << c.what() << endl;
  return Sct_SctApi::SctApiException(c.what());
  /*
  iluSctApiException *e = new iluSctApiException;
  char *buffer = new char[strlen(c.what())+1];

  cout << "Raising exception " << c.what() << endl;

  strcpy(buffer, c.what());
  buffer[strlen(c.what())] = 0;
  e->detail = buffer;
  Sct_SctApi_G::RaiseException(_status, Sct_SctApi_E_SctApiException, e);
  */
}


void SctApiServer::shutdown () {
#warning WHAT SHOULD SHUTDOWN DO IN SCTAPISERVER.  MAYBE IT SHOULD CALL SHUTDOWNALL ??
};
