#include <stdlib.h>

#include <is/isinfo.h>
#include <boost/lexical_cast.hpp>

#include <sys/types.h>
#include <sys/stat.h>

#include "configserver.h"
#include "Sct/SctNames.h"
#include "Sct/IoExceptions.h"
#include "Sct/OmniMarshalling.h"
#include "marshalling.h"

#include "Sct/SequenceUtils.h"

using namespace Sct;
using namespace SctConfiguration;
using namespace boost;

using namespace std;

// wraps a (non-corba) SctConfiguration::ConfigurationException as a 
// CORBA sctConf::ConfigurationException
sctConf::ConfigurationException ConfigServer::wrapException(::SctConfiguration::ConfigurationException& c){
  // This is extremely useful during debugging...
  cerr << "Raising exception " << c.what() << endl;
  return ::sctConf::ConfigurationException(c.what());
}

void ConfigServer::shutdown()
{
}

ConfigServer::ConfigServer(const IPCPartition& p, const std::string& name) : 
  IPCNamedObject<POA_sctConf::Configuration, ipc::single_thread>(p, name), 
  infoReceiver(new ISInfoReceiver(SctNames::getPartition()))/*, lastState(RCStateInfo::ACT_Initial) */
{
  ///This fails if sctStructure has been updated without updating the definition of ABCDModule in configuration.idl
  assert(sizeof(::sctConf::ABCDModule) == sizeof(ABCDModule));    

  // Not compatible with TApi
  IPCPartition partition("SCT");

  MRSStream mrsStream(partition);

  mrsStream << "SCT_CONFIG_START" << MRS_INFORMATION << MRS_QUALIF("sctConf") << MRS_TEXT("SCT configuration service started") << ENDM;

  std::string filename(config.getConfigurationFileName());

  ISInfoDictionary dict(partition);
  ISInfoString isFilename(filename);

  //   ISInfoString getFilename;
  //   dict.findValue("ConfigurationServer.Filename", getFilename);
  //   std::cout << "Found " << getFilename.getValue() << " in IS\n";

  if(dict.contains("ConfigurationServer.Filename")==ISInfo::Success) {
    dict.update("ConfigurationServer.Filename", isFilename);

    std::cout << "*** Filename already in IS, will run anyway but there could be another service running?\n";
    std::cout << "Replacing with " << filename << endl;
  } else {
    dict.insert("ConfigurationServer.Filename", isFilename);
  }

  // Inform configuration that it's running from IPC
  config.setIsDictionary(new ISInfoDictionary(partition));

  std::cout << "Put " << isFilename.getValue() << " into  IS\n";

}

ConfigServer::~ConfigServer() 
{
  // TApi doesn't like this, so do it here instead!
  IPCPartition partition("SCT");

  ISInfoDictionary dict(partition);

  if(dict.contains("ConfigurationServer.Filename")==ISInfo::Success) {
    dict.remove("ConfigurationServer.Filename");
  } else {
    std::cout << "*** No Filename in IS, don't need to remove it?\n";
  }
}

sctConf::PartitionList* ConfigServer::listPartitions () 
{
  try{
    list<unsigned int> parts = config.listPartitions();
    sctConf::PartitionList* result = SequenceUtils::newSafeSequence<sctConf::PartitionList>(parts.size());

    unsigned i=0;
    for(list<unsigned int>::const_iterator p = parts.begin();
	p!=parts.end();
	p++) {
      (*result)[i]=*p;
      ++i;
    }
    return result;
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

sctConf::CrateList* ConfigServer::listCratesInPartition (CORBA::ULong partition) 
{

  try{
  list<unsigned int> crates = config.listCratesInPartition(partition);
  sctConf::CrateList* result = SequenceUtils::newSafeSequence<sctConf::CrateList>(crates.size());

  unsigned i=0;
  for(list<unsigned int> :: const_iterator c = crates.begin();
      c!=crates.end();
      c++) {
    (*result)[i]=*c;
    ++i;
  }
  return result;
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

CORBA::Boolean ConfigServer::isDummyCrate (CORBA::ULong partition, CORBA::ULong crate)
{
  try{
  return config.isDummyCrate(partition, crate);
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

sctConf::RodList* ConfigServer::listRodsInCrate (CORBA::ULong partition, CORBA::ULong crate) 
{
  try{
    list<unsigned int> rods = config.listRodsInCrate(partition, crate);
    
    sctConf::RodList* result = SequenceUtils::newSafeSequence<sctConf::RodList>(rods.size());
    
    unsigned i=0;
    for(list<unsigned int> :: const_iterator r = rods.begin();
	r!=rods.end();
	r++) {
      (*result)[i]=*r;
      ++i;
    }
    return result;
  }catch(ConfigurationException& e){
    throw wrapException(e);  
  }
}

sctConf::MURList* ConfigServer::listMURSInRod (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod) 
{
try{
  list<unsigned int> MURs = config.listMURSInRod(partition, crate, rod);

  sctConf::MURList* result = SequenceUtils::newSafeSequence<sctConf::MURList>(MURs.size());

  unsigned i=0;
  for(list<unsigned int> :: const_iterator m = MURs.begin();
      m!=MURs.end();
      m++) {
    (*result)[i]=*m;
    ++i;
  }
  return result;
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

sctConf::ModuleList* ConfigServer::listModulesInMUR (CORBA::ULong partition, CORBA::ULong MUR) 
{try{
  list<string> mods = config.listModulesInMUR(partition, MUR);
  sctConf::ModuleList* result = SequenceUtils::newSafeSequence<sctConf::ModuleList>(mods.size());

  unsigned i=0;
  for(list<string> :: const_iterator m = mods.begin();
      m!=mods.end();
      m++) {
    (*result)[i]=CORBA::string_dup((*m).c_str());
    ++i;
  }
  return result;
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

sctConf::ModuleList* ConfigServer::listRModulesInMUR (CORBA::ULong partition, CORBA::ULong MUR) 
{
  try{
    list<string> mods = config.listRModulesInMUR(partition, MUR);  
    sctConf::ModuleList* result = SequenceUtils::newSafeSequence<sctConf::ModuleList>(mods.size());
    
    unsigned i=0;
    for(list<string> :: const_iterator m = mods.begin();
	m!=mods.end();
	m++) {
      (*result)[i]=CORBA::string_dup((*m).c_str());
      ++i;
    }
    return result;
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

sctConf::ModuleList* ConfigServer::listAllModules () 
{
  try{
  list<string> mods = config.listAllModules();
  
  sctConf::ModuleList* result = SequenceUtils::newSafeSequence<sctConf::ModuleList>(mods.size());

  unsigned i=0;
  for(list<string> :: const_iterator m = mods.begin();
      m!=mods.end();
      m++) {
    (*result)[i]=CORBA::string_dup((*m).c_str());
    ++i;
  }
  return result;
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

sctConf::ModuleList* ConfigServer::listUnusedModules () 
{
  try{
  list<string> mods = config.listUnusedModules();
  sctConf::ModuleList* result = SequenceUtils::newSafeSequence<sctConf::ModuleList>(mods.size());

  unsigned i=0;
  for(list<string> :: const_iterator m = mods.begin();
      m!=mods.end();
      m++) {
    (*result)[i]=CORBA::string_dup((*m).c_str());
    ++i;
  }
  return result;
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

sctConf::ABCDModule ConfigServer::getModuleConfig (const char* mod) 
{
  try{
    ABCDModule realConfig = config.getModuleConfig(mod);
    return copyABCDModuleToCorba(realConfig);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

CORBA::UShort ConfigServer::getModuleGroup (const char* mod)
{
  try{
    return config.getModuleGroup(mod);
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

void ConfigServer::setModuleGroup (const char* mod, CORBA::Short group)
{
  try{
    config.setModuleGroup(mod, group);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

sctConf::RodConfig * ConfigServer::getRodConfig (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod) 
{
  try{
  return copyRodConfigToCorba(config.getRodConfig(partition, crate, rod));
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

sctConf::SlaveConfig* ConfigServer::getDefaultSlaveConfig (CORBA::Octet type)
{
  try{
    return copySlaveConfigToCorba(config.getDefaultSlaveConfig(type));
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

sctConf::TimConfig ConfigServer::getTimConfig (CORBA::ULong partition, CORBA::ULong crate) 
{
  try{
    TimConfig myTim;
    myTim = config.getTimConfig(partition, crate);
    
    return copyTimConfigToCorba(myTim);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}
sctConf::CharBlock* ConfigServer::getFibreMappings (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod)

{
  unsigned char * mappings = 0;
  try {
    mappings = config.getFibreMappings(partition, crate, rod);
    return copyArrayToCorba<sctConf::CharBlock, unsigned char>(mappings, 48*3);
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }

}

void ConfigServer::setFibreMapping (
                                    CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                                    CORBA::ULong channel, CORBA::ULong tx, CORBA::ULong rx0, CORBA::ULong rx1)
{
  try{
    config.setFibreMapping(partition, crate, rod, channel, tx, rx0, rx1);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

sctConf::MURType ConfigServer::getMURType (CORBA::ULong MUR) 
{
  try{
  return sctConf::MURType(config.getMURType(MUR));
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

CORBA::ULong ConfigServer::getFreeMurId () 
{
  try{
    return config.getFreeMurId();
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

void ConfigServer::translateToROD (CORBA::ULong MUR, CORBA::ULong mod, 
                                   CORBA::ULong& partition, CORBA::ULong& crate, CORBA::ULong& rod, CORBA::ULong& channel) 
{
  try{
  unsigned int partitionOut, crateOut, rodOut, channelOut;
  
  config.translateToROD(MUR, mod, partitionOut, crateOut, rodOut, channelOut);
  
  partition = partitionOut;
  crate = crateOut;
  rod = rodOut;
  channel = channelOut;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateToRROD (CORBA::ULong MUR, CORBA::ULong mod, 
                                   CORBA::ULong& partition, CORBA::ULong& crate, CORBA::ULong& rod, CORBA::ULong& channel) 
{
  try{
  unsigned int partitionOut, crateOut, rodOut, channelOut;
  
  config.translateToRROD(MUR, mod, partitionOut, crateOut, rodOut, channelOut);
  
  partition = partitionOut;
  crate = crateOut;
  rod = rodOut;
  channel = channelOut;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateToSN (CORBA::ULong MUR, CORBA::ULong mod, 
                                  CORBA::String_out sn) 
{
  try{
  std::string serialNum;
  
  config.translateToSN(MUR, mod, serialNum);
  sn=CORBA::string_dup(serialNum.c_str());
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

sctConf::BOCChannelConfig ConfigServer::getBOCConfig (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong channel) 
{
  try{
  sctConf::BOCChannelConfig res;
  
  BOCChannelConfig ret = config.getBOCConfig(partition, crate, rod, channel);
  
  res.current = ret.current;
  res.delay = ret.delay;
  res.markSpace = ret.markSpace;
  
  res.threshold0 = ret.threshold0;
  res.delay0 = ret.delay0;
  
  res.threshold1 = ret.threshold1;
  res.delay1 = ret.delay1;

  return res;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

sctConf::BOCGlobalConfig ConfigServer::getBOCGlobalConfig (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod) 
{
  try{
  return copyBOCGlobalConfigToCorba(config.getBOCGlobalConfig(partition, crate, rod));
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::printModuleConfig (const sctConf::ABCDModule& conf) 
{
  try{
  config.printModuleConfig(copyCorbaToABCDModule(conf));
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::configureBOCChannel (CORBA::ULong MUR, CORBA::ULong position, const sctConf::BOCChannelConfig& conf) 
{
  try{
  BOCChannelConfig newConf; 
  newConf.current = conf.current;
  newConf.delay = conf.delay;
  newConf.markSpace = conf.markSpace;
  
  newConf.threshold0 = conf.threshold0;
  newConf.delay0 = conf.delay0;
  
  newConf.threshold1 = conf.threshold1;
  newConf.delay1 = conf.delay1;
  
  config.configureBOCChannel(MUR, position, newConf); 
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateToRMUR (CORBA::ULong RMUR, CORBA::ULong rmod, CORBA::ULong& MUR, CORBA::ULong& mod)
{
  try{
  unsigned int m, module;
  config.translateToRMUR(RMUR, rmod, m, module);
  MUR = m;
  mod = module;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateToBarrel (CORBA::ULong MUR, CORBA::ULong mod, 
                                      CORBA::ULong& barrel, CORBA::ULong& row, CORBA::Long& number) 
{
  try{
  unsigned int b, r;
  int num;
  
  config.translateToBarrel(MUR, mod, b, r, num);
  
  barrel = b;
  row = r;
  number = num;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateToEndcap (CORBA::ULong MUR, CORBA::ULong mod, 
                                      CORBA::Long& disk, CORBA::ULong& ring, CORBA::ULong& number) 
{
  try{
  int d;
  unsigned int r, n;
  config.translateToEndcap(MUR, mod, d, r, n);
  disk = d;
  ring = r;
  number = n;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateToPowerSupply (CORBA::ULong MUR, CORBA::ULong mod, CORBA::ULong& partition, CORBA::ULong& crate, CORBA::ULong& channel) 
{
  try{
    unsigned int partitionOut, crateOut, channelOut;
    
    config.translateToPowerSupply(MUR, mod, partitionOut, crateOut, channelOut);
    
    partition = partitionOut;
    crate = crateOut;
    channel = channelOut;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateFromROD (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong channel, 
                                     CORBA::ULong& MUR, CORBA::ULong& mod) 
{
  try{
  unsigned int M, module;
  config.translateFromROD(partition, crate, rod, channel, M, module);
  MUR = M;
  mod = module;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateFromRROD (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong channel, 
                                     CORBA::ULong& MUR, CORBA::ULong& mod) 
{
  try{
    unsigned int M, module;
    config.translateFromRROD(partition, crate, rod, channel, M, module);
    MUR = M;
    mod = module;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateFromSN (const char* sn, 
                                    CORBA::ULong& MUR, CORBA::ULong& mod) 
{
  try{
    unsigned int M, module;
    config.translateFromSN(std::string(sn), M, module);
    MUR = M;
    mod = module;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateFromRMUR (CORBA::ULong MUR, CORBA::ULong mod, CORBA::ULong& RMUR, CORBA::ULong& rmod)
{
  try{
    unsigned int rm, rmodule;
    config.translateFromRMUR(MUR, mod, rm, rmodule);
    RMUR = rm;
    rmod = rmodule;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateFromBarrel (CORBA::ULong barrel, CORBA::ULong row, CORBA::Long number, 
                                        CORBA::ULong& MUR, CORBA::ULong& mod) 
{
  try{
    unsigned int M, module;
    config.translateFromBarrel(barrel, row, number, M, module);
    MUR = M;
    mod = module;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateFromEndcap (CORBA::Long disk, CORBA::ULong ring, CORBA::ULong number, 
                                        CORBA::ULong& MUR, CORBA::ULong& mod) 
{
  try{
    unsigned int M, module;
    config.translateFromEndcap(disk, ring, number, M, module);
    MUR = M;
    mod = module;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::translateFromPowerSupply (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong channel, CORBA::ULong& MUR, CORBA::ULong& mod) 
{
  
  try{
    unsigned int M, module;
    config.translateFromPowerSupply(partition, crate, channel, M, module);
    MUR = M;
    mod = module;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::configureModuleFromStructure (const char* mod, const sctConf::ABCDModule& conf)
{
  try{
  const ABCDModule& modConfig = copyCorbaToABCDModule(conf);
  
  config.configureModuleFromStructure(mod, modConfig);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::configureModuleFromFile (const char* filename)
{  
  try{
    config.configureModuleFromFile(filename);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::clearAll ()
{
  try{
    config.clearAll();
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::namePartition (CORBA::ULong partition, const char* name)
{
  try{
    config.namePartition(partition, name);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

// void ConfigServer::configureROD (
//                                  CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
//                                  sctConf::RodConfig * conf) 
// {
//   cout << "Can't update ROD config\n";
//   //  config.updateRodConfig(partition, crate, rod, conf);
// }

void ConfigServer::loadConfiguration (const char* filename) 
{  
  try{
    config.loadConfiguration(filename);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::saveConfiguration (const char* filename) 
{ 
  try{
    config.saveConfiguration(filename);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::saveModuleConfiguration (const char* sn, const char* filename) 
{
  try{
    config.saveModuleConfiguration(sn, filename);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

char* ConfigServer::getModuleConfigurationString (const char* modsn, const sctConf::ABCDModule& mConf)
{
  try{
  char* result = 0;

  
  string buffer = config.getModuleConfigurationString(modsn, copyCorbaToABCDModule(mConf));
  
  result = new char[buffer.length() + 1];
  strncpy(result, buffer.c_str(), buffer.length() + 1);
  return result;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::writePowerSupplyConfiguration (const char* filename)
{
  try{
    config.writePowerSupplyConfiguration(filename);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::mapRODMUR (
                              CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                              CORBA::ULong order, CORBA::ULong number) 
{ 
  try{
    config.mapRODMUR(partition, crate, rod, order, number);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::unmapRODMUR (
                              CORBA::ULong MUR) 
{ 
  try{
    config.unmapRODMUR(MUR);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::mapBarrelMUR (
                                 CORBA::ULong MUR, 
                                 CORBA::ULong barrel, CORBA::ULong row, CORBA::Long position)
{ 
  try{ 
    config.mapBarrelMUR(MUR, barrel, row, position); 
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::unmapBarrelMUR (CORBA::ULong MUR)
{
  try{
    config.unmapBarrelMUR(MUR);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::mapEndcapMUR (
                                 CORBA::ULong MUR, 
                                 CORBA::Long disk, CORBA::ULong quadrant, CORBA::ULong position)
{
  try{
    config.mapEndcapMUR(MUR, disk, quadrant, position);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::unmapEndcapMUR (CORBA::ULong MUR)
{
  try{
    config.unmapEndcapMUR(MUR); 
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::mapPowerChannel (CORBA::ULong MUR, CORBA::ULong number, CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong channel)
{
  try{
    config.mapPowerChannel(MUR, number, partition, crate, channel);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::unmapPowerChannel (CORBA::ULong MUR, CORBA::ULong number)
{ 
  try{
    config.unmapPowerChannel(MUR, number);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::mapModuleMUR (
                                 CORBA::ULong MUR, CORBA::ULong order, CORBA::ULong RMUR, CORBA::ULong rorder, 
                                 const char* number) 
{
  try{
    config.mapModuleMUR(MUR, order, RMUR, rorder, number);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::unmapModuleMUR (CORBA::ULong MUR, CORBA::ULong number)
{
     config.unmapModuleMUR(MUR, number);
}

void ConfigServer::getMapMURROD (CORBA::ULong MUR, CORBA::ULong& partition, CORBA::ULong& crate, CORBA::ULong& rod, CORBA::ULong& order)
{
  try{
  unsigned int partitionOut, crateOut, rodOut, orderOut;  
  config.getMapMURROD(MUR, partitionOut, crateOut, rodOut, orderOut);
  partition = partitionOut;
  crate = crateOut;
  rod = rodOut;
  order = orderOut;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::getMapRODMUR (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, CORBA::ULong order, CORBA::ULong& MUR)
{ 
  try{
  unsigned int MURout;
  config.getMapRODMUR(partition, crate, rod, order, MURout);
  MUR = MURout;
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

sctConf::MURType ConfigServer::getMapMURPhysical (CORBA::ULong MUR, CORBA::ULong& plane, CORBA::ULong& section, CORBA::ULong& position)
{ 
  try{
  MURType type = UNKNOWN;
  
  int p, s, pos;
  config.getMapMURPhysical(MUR, type, p, s, pos);
  plane = p;
  section = s;
  position = pos;
  
  return sctConf::MURType(type);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::swapMURNames (
                                 CORBA::ULong MUR1, CORBA::ULong MUR2) 
{
  try{
  config.swapMURNames(MUR1, MUR2);
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

  void ConfigServer::modifyDefaultPowerParam(const char* state, const char* name, const char* type, CORBA::Float value) {
    try{
      config.modifyDefaultPowerParam(state, name, type, value);
    }catch(ConfigurationException& e){
      throw wrapException(e);
    };
  };
  void ConfigServer::modifyHVCardParam(CORBA::ULong crate, CORBA::ULong card, const char* name, const char* type, CORBA::Float value) {
    try{
      config.modifyHVCardParam(crate, card, name, type, value);
    }catch(ConfigurationException& e){
      throw wrapException(e);
    };
  };
  void ConfigServer::modifyLVCardParam(CORBA::ULong crate, CORBA::ULong card, const char* name, const char* type, CORBA::Float value) {
    try{
      config.modifyLVCardParam(crate, card, name, type, value);
    }catch(ConfigurationException& e){
      throw wrapException(e);
    };
  };
  void ConfigServer::modifyPowerParam(CORBA::ULong MUR, CORBA::ULong number, const char* state, const char* name, const char* type, CORBA::Float value) {
    try{
      config.modifyPowerParam(MUR, number, state, name, type, value);
    }catch(ConfigurationException& e){
      throw wrapException(e);
    };
  };
  void ConfigServer::modifyPowerParamCC(CORBA::ULong crate, CORBA::ULong channel, const char* state, const char* name, const char* type, CORBA::Float value) {
    try{
      config.modifyPowerParamCC(crate, channel, state, name, type, value);
    }catch(ConfigurationException& e){
      throw wrapException(e);
    };
  };
  CORBA::Float ConfigServer::getDefaultPowerParam(const char* state, const char* name, const char* type) {
    try{
      const CORBA::Float result = config.getDefaultPowerParam(state, name, type);
      return result;
    }catch(ConfigurationException& e){
      throw wrapException(e); 
    };
  }
  CORBA::Float ConfigServer::getHVCardParam(CORBA::ULong crate, CORBA::ULong card, const char* name, const char* type) {
    try{
      const CORBA::Float result = config.getHVCardParam(crate, card, name, type);
      return result;
    }catch(ConfigurationException& e){
      throw wrapException(e); 
    };
  }
  CORBA::Float ConfigServer::getLVCardParam(CORBA::ULong crate, CORBA::ULong card, const char* name, const char* type) {
    try{
      const CORBA::Float result = config.getHVCardParam(crate, card, name, type);
      return result;
    }catch(ConfigurationException& e){
      throw wrapException(e); 
    };
  }
  CORBA::Float ConfigServer::getPowerParam(CORBA::ULong MUR, CORBA::ULong number, const char* state, const char* name, const char* type) {
    try{
      const CORBA::Float result = config.getPowerParam(MUR, number, state, name, type);
      return result;
    }catch(ConfigurationException& e){
      throw wrapException(e); 
    };
  }
  CORBA::Float ConfigServer::getPowerParamCC(CORBA::ULong crate, CORBA::ULong channel, const char* state, const char* name, const char* type) {
    try{
      const CORBA::Float result = config.getPowerParamCC(crate, channel, state, name, type);
      return result;
    }catch(ConfigurationException& e){
      throw wrapException(e); 
    };
  }

sctConf::NameList* ConfigServer::listDefaultPowerParams ()
{
  try{
    return copyStringListToCorba<sctConf::NameList>(config.listDefaultPowerParams()); 
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

CORBA::Float ConfigServer::getCratePowerParam (CORBA::ULong crate, const char* name)
{
  try{
  return config.getCratePowerParam(crate, name); 
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

sctConf::NameList* ConfigServer::listDCSChannelParams ()
{
  try{
  return copyStringListToCorba<sctConf::NameList>(config.listDCSChannelParams()); 
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

sctConf::NameList* ConfigServer::listDCSCardParams ()
{
  try{
    return copyStringListToCorba<sctConf::NameList>(config.listDCSCardParams());
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  } 
}

sctConf::NameList* ConfigServer::listDCSCrateParams ()
{
  try{
  return copyStringListToCorba<sctConf::NameList>(config.listDCSCrateParams());
  }catch(ConfigurationException& e){
    throw wrapException(e); 
  }
}

void ConfigServer::configureROD (
                                 CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                                 const sctConf::RodConfig& conf) 
{
  try{
  config.configureROD(partition, crate, rod, copyCorbaToRodConfig(conf));
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::configureTIM (CORBA::ULong partition, CORBA::ULong crate, 
                                 const sctConf::TimConfig& conf) 
{
  try{
  config.configureTIM(partition, crate, copyCorbaToTimConfig(conf));
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}

void ConfigServer::configureBOC (CORBA::ULong partition, CORBA::ULong crate, CORBA::ULong rod, 
                                 const sctConf::BOCGlobalConfig& conf) 
{
  try{
  config.configureBOC(partition, crate, rod, copyCorbaToBOCGlobalConfig(conf));
  }catch(ConfigurationException& e){
    throw wrapException(e);
  }
}
